import datetime
import http
import json
from abc import abstractmethod
from typing import Union, List, Optional, Tuple

import requests

from ravendb.primitives import constants
from ravendb.http.misc import AggressiveCacheOptions, AggressiveCacheMode
from ravendb.http.request_executor import RequestExecutor
from ravendb.extensions.http_extensions import HttpExtensions
from ravendb.http.http_cache import HttpCache, ReleaseCacheItem
from ravendb.http.raven_command import RavenCommand, RavenCommandResponseType
from ravendb.http.server_node import ServerNode
from ravendb.tools.utils import CaseInsensitiveDict


class Content:
    @abstractmethod
    def write_content(self) -> dict:
        pass


class GetRequest:
    def __init__(self):
        self.headers = CaseInsensitiveDict()
        self.url: Union[None, str] = None
        self.query: Union[None, str] = None
        self.method: Union[None, str] = None
        self.content: Union[None, Content] = None
        self.can_cache_aggressively = True

    @property
    def url_and_query(self) -> str:
        if not self.query:
            return self.url
        if self.query.startswith("?"):
            return self.url + self.query
        return self.url + "?" + self.query


class GetResponse:
    def __init__(self):
        self.headers = CaseInsensitiveDict()
        self.elapsed: Union[None, datetime.timedelta] = None
        self.result: Union[None, str] = None
        self.status_code: Union[None, int] = None
        self.force_retry: Union[None, bool] = None

    @property
    def request_has_errors(self):
        return self.status_code not in [
            0,
            http.HTTPStatus.OK,
            http.HTTPStatus.CREATED,
            http.HTTPStatus.NON_AUTHORITATIVE_INFORMATION,
            http.HTTPStatus.NO_CONTENT,
            http.HTTPStatus.NOT_MODIFIED,
            http.HTTPStatus.NOT_FOUND,
        ]


class Cached:
    def __init__(self, size: int):
        self.__size = size
        self.values: Union[None, List[Tuple[ReleaseCacheItem, str]]] = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if not self.values:
            return

        for value in self.values:
            if value is not None:
                value[0].close()

        self.values = None


class MultiGetCommand(RavenCommand):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_cache()

    def __init__(self, request_executor: RequestExecutor, commands: List[GetRequest]):
        super(MultiGetCommand, self).__init__(list)

        if not request_executor:
            raise ValueError("RequestExecutor cannot be None")

        cache = request_executor.cache

        if cache is None:
            raise ValueError("Cache cannot be None")

        if commands is None:
            raise ValueError("Command cannot be None")

        self.__request_executor = request_executor
        self.__http_cache: HttpCache = request_executor.cache
        self.__commands = commands
        self._response_type = RavenCommandResponseType.RAW
        self.aggressively_cached: Union[None, bool] = None
        self.__base_url: str = ""
        self.__cached: Union[None, Cached] = None

    def create_request(self, node: ServerNode) -> Optional[requests.Request]:
        self.__base_url = f"{node.url}/databases/{node.database}"
        url = self.__base_url + "/multi_get"
        request = requests.Request("POST", url)

        request.data = {
            "Requests": [
                {
                    "Url": f"/databases/{node.database}{command.url}",
                    "Query": command.query,
                    "Method": command.method,
                    "Headers": command.headers,
                    "Content": command.content.write_content() if command.content else None,
                }
                for command in self.__commands
            ]
        }

        return request

    def __maybe_read_all_from_cache(self, options: AggressiveCacheOptions):
        self.close_cache()

        read_all_from_cache = options is not None
        track_changes = read_all_from_cache and options.mode == AggressiveCacheMode.TRACK_CHANGES
        for command in self.__commands:
            cache_key = self.__get_cache_key(command)[0]
            cached_item, change_vector, cached_ref = self.__http_cache.get(cache_key, "", "")
            cached_item: ReleaseCacheItem
            if cached_item.item is None:
                try:
                    read_all_from_cache = False
                    continue
                finally:
                    cached_item.close()

            if read_all_from_cache and (
                track_changes
                and cached_item.might_have_been_modified
                or cached_item.age > options.duration
                or not command.can_cache_aggressively
            ):
                read_all_from_cache = False

            command.headers[constants.Headers.IF_NONE_MATCH] = change_vector
            if self.__cached is None:
                self.__cached = Cached(len(self.__commands))

            self.__cached.values[self.__commands.index(command)] = (cached_item, cached_ref)

        if read_all_from_cache:
            with self.__cached as context:
                self.result = []
                for i in range(len(self.__commands)):
                    item_and_cached = self.__cached.values[i]
                    get_response = GetResponse()
                    get_response.result = item_and_cached[1]
                    get_response.status_code = http.HTTPStatus.NOT_MODIFIED

                    self.result.append(get_response)

            self.__cached = None

        return read_all_from_cache

    def __get_cache_key(self, command: GetRequest) -> (str, str):
        req_url = self.__base_url + command.url_and_query
        return command.method + "-" + req_url if command.method else req_url, req_url

    def set_response_raw(self, response: requests.Response, stream: bytes) -> None:
        try:
            try:
                response_temp = response.json()
                if "Results" not in response_temp:
                    self._throw_invalid_response()

                i = 0
                self.result = []

                for get_response in self.read_responses(response_temp):
                    command = self.__commands[i]
                    self.__maybe_set_cache(get_response, command, i)

                    if self.__cached is not None and get_response.status_code == http.HTTPStatus.NOT_MODIFIED:
                        cloned_response = GetResponse()
                        cloned_response.result = self.__cached.values[i][1]
                        cloned_response.status_code = http.HTTPStatus.NOT_MODIFIED
                        self.result.append(cloned_response)
                    else:
                        self.result.append(get_response)

                    i += 1
            finally:
                if self.__cached is not None:
                    self.__cached.close()
        except Exception as e:
            self._throw_invalid_response(e)

    def __maybe_set_cache(self, get_response: GetResponse, command: GetRequest, cached_index: int):
        if get_response.status_code == http.HTTPStatus.NOT_MODIFIED:
            # if not modified - update age
            if self.__cached is not None:
                self.__cached.values[cached_index][0].not_modified()
            return

        cache_key = self.__get_cache_key(command)[0]

        result = get_response.result
        if result == None:
            self.__http_cache.set_not_found(cache_key, self.aggressively_cached)
            return

        change_vector = HttpExtensions.get_etag_header(get_response.headers)
        if change_vector is None:
            return

        self.__http_cache.set(cache_key, change_vector, result)

    @staticmethod
    def read_responses(response_json: dict) -> List[GetResponse]:
        responses = []
        for response in response_json["Results"]:
            responses.append(MultiGetCommand.read_response(response))
        return responses

    @staticmethod
    def read_response(response_json: dict) -> GetResponse:
        get_response = GetResponse()
        get_response.result = json.dumps(response_json["Result"])  # todo: perf - redundant dump after response.json()
        get_response.headers = CaseInsensitiveDict(response_json["Headers"])
        if response_json["StatusCode"] == -1:
            MultiGetCommand._throw_invalid_response()
        get_response.status_code = response_json["StatusCode"]

        return get_response

    @property
    def is_read_request(self) -> bool:
        return False

    def close_cache(self):
        # If _cached is not null - it means that the client approached with this multitask request to node
        # and the request failed and now client tries to send it to another node.

        if self.__cached is not None:
            self.__cached.close()

        self.__cached = None
        # The client sends the commands.
        # Some of which could be saved in cache with a response
        # that includes the change vector that received from the old fallen node.
        # The client can't use those responses because their URLs are different
        # (include the IP and port of the old node), because of that the client
        # needs to get those docs again from the new node.

        for command in self.__commands:
            command.headers.remove(constants.Headers.IF_NONE_MATCH)



from ravendb.documents.commands.batches import (
    BatchOptions,
    DeleteAttachmentCommandData,
    PatchCommandData,
    PutAttachmentCommandData,
    CommandData,
    CopyAttachmentCommandData,
    MoveAttachmentCommandData,
    BatchPatchCommandData,
    CountersBatchCommandData,
    PutCompareExchangeCommandData,
    DeleteCompareExchangeCommandData,
)
from ravendb.documents.commands.crud import DeleteDocumentCommand, PutDocumentCommand
from ravendb.documents.indexes.analysis.definitions import AnalyzerDefinition
from ravendb.documents.indexes.definitions import (
    IndexDeploymentMode,
    IndexDefinition,
    AbstractCommonApiForIndexes,
    AdditionalAssembly,
    IndexFieldOptions,
    RollingIndex,
    RollingIndexDeployment,
    RollingIndexState,
    IndexSourceType,
    AutoIndexDefinition,
    AutoIndexFieldOptions,
)
from ravendb.documents.indexes.abstract_index_creation_tasks import (
    AbstractIndexDefinitionBuilder,
    AbstractIndexCreationTask,
)
from ravendb.documents.indexes.spatial.configuration import AutoSpatialOptions
from ravendb.documents.operations.attachments import (
    DeleteAttachmentOperation,
    PutAttachmentOperation,
    GetAttachmentOperation,
    AttachmentRequest,
)
from ravendb.documents.operations.backups.settings import (
    BackupConfiguration,
    AmazonSettings,
    AzureSettings,
    FtpSettings,
    GlacierSettings,
    LocalSettings,
    PeriodicBackupConfiguration,
    S3Settings,
    BackupSettings,
    BackupStatus,
    GoogleCloudSettings,
)
from ravendb.documents.operations.batch import BatchOperation
from ravendb.documents.operations.compare_exchange.compare_exchange import (
    CompareExchangeValue,
    CompareExchangeSessionValue,
    CompareExchangeValueState,
)
from ravendb.documents.operations.compare_exchange.compare_exchange_value_result_parser import (
    CompareExchangeValueResultParser,
)
from ravendb.documents.operations.compare_exchange.operations import (
    PutCompareExchangeValueOperation,
    GetCompareExchangeValueOperation,
    CompareExchangeResult,
    GetCompareExchangeValuesOperation,
    DeleteCompareExchangeValueOperation,
)
from ravendb.documents.operations.configuration.operations import (
    GetServerWideClientConfigurationOperation,
    PutServerWideClientConfigurationOperation,
    ClientConfiguration,
    GetClientConfigurationOperation,
    PutClientConfigurationOperation,
)
from ravendb.documents.operations.configuration.definitions import StudioConfiguration, StudioEnvironment

from ravendb.documents.operations.connection_strings import ConnectionString
from ravendb.documents.operations.etl.configuration import EtlConfiguration, RavenEtlConfiguration
from ravendb.documents.operations.etl.olap import OlapEtlConfiguration
from ravendb.documents.operations.etl.sql import SqlEtlConfiguration
from ravendb.documents.operations.executor import MaintenanceOperationExecutor, SessionOperationExecutor
from ravendb.documents.operations.expiration.configuration import ExpirationConfiguration
from ravendb.documents.operations.indexes import (
    GetIndexNamesOperation,
    DisableIndexOperation,
    EnableIndexOperation,
    GetIndexingStatusOperation,
    GetIndexesStatisticsOperation,
    GetIndexStatisticsOperation,
    GetIndexesOperation,
    GetTermsOperation,
    IndexHasChangedOperation,
    PutIndexesOperation,
    StopIndexingOperation,
    StartIndexingOperation,
    StopIndexOperation,
    StartIndexOperation,
    DeleteIndexOperation,
    SetIndexesLockOperation,
    SetIndexesPriorityOperation,
    GetIndexOperation,
    GetIndexErrorsOperation,
    IndexingStatus,
)
from ravendb.documents.operations.lazy.definition import LazyOperation
from ravendb.documents.operations.misc import DeleteByQueryOperation, GetOperationStateOperation, QueryOperationOptions
from ravendb.documents.operations.patch import (
    PatchOperation,
    PatchByQueryOperation,
    PatchRequest,
    PatchResult,
    PatchStatus,
)
from ravendb.documents.operations.refresh.configuration import RefreshConfiguration
from ravendb.documents.operations.replication.definitions import (
    ExternalReplication,
    PullReplicationAsSink,
    PullReplicationDefinition,
    ReplicationNode,
    ExternalReplicationBase,
)
from ravendb.documents.operations.revisions import (
    RevisionsCollectionConfiguration,
    RevisionsConfiguration,
)
from ravendb.documents.operations.statistics import (
    GetCollectionStatisticsOperation,
    CollectionStatistics,
    GetStatisticsOperation,
    DatabaseStatistics,
    IndexInformation,
    GetDetailedStatisticsOperation,
    DetailedDatabaseStatistics,
)
from ravendb.documents.queries.explanation import ExplanationOptions, Explanations
from ravendb.documents.queries.facets.builders import RangeBuilder, FacetBuilder, FacetOperations
from ravendb.documents.queries.facets.definitions import (
    FacetAggregationField,
    Facet,
    RangeFacet,
    FacetBase,
    GenericRangeFacet,
)
from ravendb.documents.queries.facets.queries import (
    AggregationRawDocumentQuery,
    AggregationDocumentQuery,
    AggregationQueryBase,
)
from ravendb.documents.queries.group_by import GroupBy, GroupByMethod
from ravendb.documents.queries.highlighting import HighlightingOptions, QueryHighlightings
from ravendb.documents.queries.index_query import IndexQuery
from ravendb.documents.queries.misc import SearchOperator
from ravendb.documents.queries.more_like_this import (
    MoreLikeThisOperations,
    MoreLikeThisBase,
    MoreLikeThisBuilder,
    MoreLikeThisOptions,
)
from ravendb.documents.queries.query import QueryOperator, ProjectionBehavior, QueryData, QueryResult, QueryTimings
from ravendb.documents.queries.sorting import SorterDefinition
from ravendb.documents.queries.spatial import (
    SpatialCriteriaFactory,
    SpatialCriteria,
    CircleCriteria,
    DynamicSpatialField,
    WktCriteria,
    PointField,
)
from ravendb.documents.queries.suggestions import (
    SuggestionBuilder,
    SuggestionDocumentQuery,
    StringDistanceTypes,
    SuggestionOptions,
    SuggestionBase,
    SuggestionResult,
    SuggestionSortMode,
)
from ravendb.documents.session.cluster_transaction_operation import ClusterTransactionOperations
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.documents.session.document_session import DocumentSession
from ravendb.documents.session.entity_to_json import EntityToJson
from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
    InMemoryDocumentSessionOperations,
)
from ravendb.documents.session.loaders.include import IncludeBuilder, IncludeBuilderBase, QueryIncludeBuilder
from ravendb.documents.session.loaders.loaders import (
    LoaderWithInclude,
    LazyMultiLoaderWithInclude,
    MultiLoaderWithInclude,
)
from ravendb.documents.session.misc import (
    CmpXchg,
    DocumentsChanges,
    ForceRevisionStrategy,
    MethodCall,
    OrderingType,
    JavaScriptMap,
    DocumentQueryCustomization,
    ResponseTimeInformation,
    TransactionMode,
    SessionOptions,
)
from ravendb.documents.session.operations.lazy import (
    LazySessionOperations,
    LazyAggregationQueryOperation,
    LazyLoadOperation,
    LazyQueryOperation,
    LazyStartsWithOperation,
    LazySuggestionQueryOperation,
    LazyConditionalLoadOperation,
)
from ravendb.documents.session.operations.load_operation import LoadOperation
from ravendb.documents.session.operations.operations import LoadStartingWithOperation, MultiGetOperation
from ravendb.documents.session.operations.query import QueryOperation
from ravendb.documents.session.query import (
    AbstractDocumentQuery,
    DocumentQuery,
    RawDocumentQuery,
    QueryStatistics,
    WhereParams,
)
from ravendb.documents.session.query_group_by import GroupByDocumentQuery, GroupByField
from ravendb.documents.session.utils.document_query import DocumentQueryHelper
from ravendb.documents.session.utils.includes_util import IncludesUtil
from ravendb.documents.store.definition import DocumentStore, DocumentStoreBase
from ravendb.documents.store.lazy import Lazy
from ravendb.documents.session.conditional_load import ConditionalLoadResult
from ravendb.documents.store.misc import IdTypeAndName
from ravendb.http.misc import AggressiveCacheOptions, Broadcast, LoadBalanceBehavior, ReadBalanceBehavior
from ravendb.http.raven_command import RavenCommand
from ravendb.http.request_executor import ClusterRequestExecutor, RequestExecutor
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import (
    ClusterTopology,
    CurrentIndexAndNode,
    CurrentIndexAndNodeAndEtag,
    RaftCommand,
    NodeSelector,
    Topology,
    UpdateTopologyParameters,
)

# StatusCode
# UriUtility
# ServerWide
# CompactSettings
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.json.result import BatchCommandResult
from ravendb.serverwide.commands import GetDatabaseTopologyCommand, GetClusterTopologyCommand
from ravendb.serverwide.misc import DocumentsCompressionConfiguration, DeletionInProgressStatus

# IDatabaseTaskStatus
from ravendb.serverwide.operations.certificates import (
    CertificateMetadata,
    EditClientCertificateOperation,
    ReplaceClusterCertificateOperation,
    GetCertificateMetadataOperation,
    GetCertificatesMetadataOperation,
    CertificateDefinition,
    CertificateRawData,
    CreateClientCertificateOperation,
    DatabaseAccess,
    DeleteCertificateOperation,
    GetCertificateOperation,
    GetCertificatesOperation,
    GetCertificatesResponse,
    PutClientCertificateOperation,
    SecurityClearance,
)
from ravendb.serverwide.operations.common import (
    BuildNumber,
    GetBuildNumberOperation,
    GetDatabaseNamesOperation,
    GetServerWideOperationStateOperation,
    CreateDatabaseOperation,
    GetDatabaseRecordOperation,
)

from ravendb.documents.identity.hilo import (
    HiLoIdGenerator,
    MultiTypeHiLoGenerator,
    MultiDatabaseHiLoGenerator,
    HiLoResult,
    GenerateEntityIdOnTheClient,
)

# todo: Serverwide
# ReorderDatabaseMembersOperation
# ConfigureRevisionsForConflictsOperation
# UpdateDatabaseOperation
# GetServerWideBackupConfigurationOperation
# SetDatabaseDynamicDistributionOperation
# UpdateUnusedDatabasesOperation

# todo: Serverwide Operations
# Operations
# DeleteDatabasesOperation
# ServerWideOperationCompletionAwaiter
# GetLogsConfigurationResult
# GetLogsConfigurationOperation
# LogMode
# SetLogsConfigurationOperation
# DeleteServerWideBackupConfigurationOperation
# GetServerWideBackupConfigurationsOperation
# PutServerWideBackupConfigurationOperation
# ServerWideBackupConfiguration
# DatabaseSettings
# GetDatabaseSettingsOperation
# PutDatabaseSettingsOperation
# GetTcpInfoCommand
# AddClusterNodeCommand
# ServerWide
# ModifyConflictSolverOperation

# todo: Operations and Commands
# BulkInsertOperation
# CollectionDetails
# BackupTaskType
# DatabaseHealthCheckOperation
# DetailedCollectionStatistics
# GetDetailedCollectionStatisticsOperation
# OperationAbstractions
# CompactDatabaseOperation
# PutConnectionStringOperation
# DeleteSorterOperation
# PutSortersOperation
# CompareExchangeValueJsonConverter
# ICompareExchangeValue
# GetServerWideExternalReplicationsResponse
# GetNextOperationIdCommand
# KillOperationCommand
# NextIdentityForCommand
# SeedIdentityForCommand
# ExplainQueryCommand
# GetIdentitiesOperation
# OperationCompletionAwaiter
# DeleteIndexErrorsOperation
# ResetIndexOperation
# GetServerWideBackupConfigurationsResponse
# NextIdentityForOperation
# SeedIdentityForOperation
# IOperationProgress
# IOperationResult
# PullReplicationDefinitionAndCurrentConnections
# DetailedReplicationHubAccess
# GetReplicationHubAccessOperation
# PreventDeletionsMode
# PullReplicationMode
# RegisterReplicationHubAccessOperation
# ReplicationHubAccess
# ReplicationHubAccessResult
# ReplicationHubAccessResponse
# UnregisterReplicationHubAccessOperation
# UpdatePullReplicationAsSinkOperation
# GetConflictsCommand
# PutAttachmentCommandHelper
# SetupDocumentBase
# StreamResultResponse
# StreamResult
# GetRevisionOperation
# GetRevisionsCountOperation
# IEagerSessionOperations
# LazyClusterTransactionOperations
# LazyGetCompareExchangeValueOperation
# LazyGetCompareExchangeValuesOperation
# LazyRevisionOperation
# LazyRevisionOperations
# StreamOperation
# ConfigureRevisionsOperation
# GetRevisionsOperation
# RevisionsResult
# GetConnectionStringsOperation
# RemoveConnectionStringOperation
# SqlEtlTable
# OlapEtlFileFormat
# OlapEtlTable
# Transformation
# AddEtlOperation
# UpdateEtlOperation
# ResetEtlOperation
# DisableDatabaseToggleResult
# ConfigureExpirationOperation
# DeleteOngoingTaskOperation
# GetPullReplicationHubTasksInfoOperation
# OngoingTaskPullReplicationAsSink
# OngoingTaskPullReplicationAsHub
# OngoingTaskType
# RunningBackup
# NextBackup
# GetOngoingTaskInfoOperation
# ToggleOngoingTaskStateOperation
# ConfigureRefreshOperation
# ConfigureRefreshOperationResult
# ToggleDatabasesStateOperation
# StartTransactionsRecordingOperation
# StopTransactionsRecordingOperation

# todo: backup
# BackupEncryptionSettings
# BackupEncryptionSettings
# GetPeriodicBackupStatusOperation
# GetPeriodicBackupStatusOperationResult
# LastRaftIndex
# PeriodicBackupStatus
# RestoreBackupConfiguration
# RestoreBackupOperation
# StartBackupOperation
# StartBackupOperationResult
# UpdatePeriodicBackupOperation
# UpdatePeriodicBackupOperationResult
# UploadProgress
# UploadState
# CompressionLevel
# GetBackupConfigurationScript
# RestoreBackupConfigurationBase
# RestoreFromAzureConfiguration
# RestoreFromGoogleCloudConfiguration
# RestoreFromS3Configuration
# RestoreType
# RetentionPolicy

# todo: Indexes
# Enums
# IndexDefinitionHelper
# IndexStats
# Indexes
# IndexDefinitionBase
# AbstractCsharpIndexCreationTask
# AbstractCsharpMultiMapIndexCreationTask
# AbstractJavaScriptIndexCreationTask
# AbstractJavaScriptMultiMapIndexCreationTask
# AbstractRawJavaScriptIndexCreationTask
# AbstractCountersIndexCreationTask
# AbstractGenericCountersIndexCreationTask
# AbstractCsharpCountersIndexCreationTask
# AbstractMultiMapCountersIndexCreationTask
# AbstractRawJavaScriptCountersIndexCreationTask
# CountersIndexDefinition
# CountersIndexDefinitionBuilder
# AbstractGenericTimeSeriesIndexCreationTask
# AbstractMultiMapTimeSeriesIndexCreationTask
# AbstractCsharpTimeSeriesIndexCreationTask
# AbstractRawJavaScriptTimeSeriesIndexCreationTask
# AbstractTimeSeriesIndexCreationTask
# TimeSeriesIndexDefinition
# TimeSeriesIndexDefinitionBuilder

# todo: Store
# DocumentAbstractions

# todo: Subscriptions
# SubscriptionBatch
# DocumentSubscriptions
# SubscriptionWorker
# SubscriptionWorkerOptions
# SubscriptionCreationOptions
# Revision
# SubscriptionState
# SubscriptionCreationOptions
# UpdateSubscriptionResult
# SubscriptionOpeningStrategy
# SubscriptionUpdateOptions

# todo: Session
# IAbstractDocumentQueryImpl
# ILazyRevisionsOperations
# IAdvancedSessionOperations
# IDocumentQueryBuilder
# IDocumentQueryBaseSingle
# IEnumerableQuery
# IFilterDocumentQueryBase
# IGraphDocumentQuery
# IGroupByDocumentQuery
# IQueryBase
# QueryEvents
# QueryOptions
# StreamQueryStatistics
# SessionEvents
# ILazyClusterTransactionOperations
# ISessionDocumentAppendTimeSeriesBase
# ISessionDocumentDeleteTimeSeriesBase
# ISessionDocumentRollupTypedAppendTimeSeriesBase
# ISessionDocumentRollupTypedTimeSeries
# ISessionDocumentTimeSeries
# ISessionDocumentTypedAppendTimeSeriesBase
# ISessionDocumentTypedTimeSeries
# DocumentResultStream
# SessionDocumentRollupTypedTimeSeries
# SessionDocumentTimeSeries
# SessionDocumentTypedTimeSeries
# SessionTimeSeriesBase
# ICounterIncludeBuilder
# IAbstractTimeSeriesIncludeBuilder
# ICompareExchangeValueIncludeBuilder
# IDocumentIncludeBuilder
# IGenericIncludeBuilder
# IGenericRevisionIncludeBuilder
# IGenericTimeSeriesIncludeBuilder
# ISubscriptionIncludeBuilder
# ISubscriptionTimeSeriesIncludeBuilder
# TimeSeriesIncludeBuilder
# SubscriptionIncludeBuilder:
# ILazyLoaderWithInclude
# ITimeSeriesIncludeBuilder
# DocumentSessionAttachments
# DocumentSessionAttachmentsBase
# DocumentSessionRevisions
# DocumentSessionRevisionsBase
# IAttachmentsSessionOperations
# IRevisionsSessionOperations
# MetadataObject
# ISessionDocumentCounters
# CounterInternalTypes
# SessionDocumentCounters
# TimeSeriesEntry
# TimeSeriesValue
# TimeSeriesValuesHelper
# TypedTimeSeriesEntry
# TypedTimeSeriesRollupEntry
# TimeSeriesOperations

# todo: Batch
# StreamResult

# todo: Counters
# CounterBatch
# GetCountersOperation
# CounterBatchOperation
# CounterOperationType
# CounterOperation
# DocumentCountersOperation
# CounterDetail
# CountersDetail

# todo: TimeSeries
# AggregationType
# RawTimeSeriesTypes
# ConfigureRawTimeSeriesPolicyOperation
# ConfigureTimeSeriesOperation
# ConfigureTimeSeriesOperationResult
# ConfigureTimeSeriesPolicyOperation
# ConfigureTimeSeriesValueNamesOperation
# GetMultipleTimeSeriesOperation
# GetTimeSeriesOperation
# GetTimeSeriesStatisticsOperation
# RawTimeSeriesPolicy
# RemoveTimeSeriesPolicyOperation
# TimeSeriesBatchOperation
# TimeSeriesCollectionConfiguration
# TimeSeriesConfiguration
# TimeSeriesDetails
# TimeSeriesItemDetail
# TimeSeriesOperation
# TimeSeriesPolicy
# TimeSeriesRange
# TimeSeriesCountRange
# TimeSeriesRangeType
# TimeSeriesTimeRange
# TimeSeriesRangeResult
# TimeSeriesStatistics
# AbstractTimeSeriesRange

# todo: Auth
# AuthOptions

# todo: Types
# Callbacks
# Contracts
# Types

# todo: Queries
# WktField
# FacetSetup
# Facets
# HighlightingParameters
# Hightlightings
# ITimeSeriesQueryBuilder
# TimeSeriesAggregationResult
# TimeSeriesQueryBuilder
# TimeSeriesQueryResult
# TimeSeriesRangeAggregation
# TimeSeriesRawResult
# TypedTimeSeriesAggregationResult
# TypedTimeSeriesRangeAggregation
# TypedTimeSeriesRawResult

# todo: More Like This
# IMoreLikeThisBuilderBase
# MoreLikeThisStopWords

# todo: Suggestions
# ISuggestionOperations

# todo: Attachments
# Attachments

# todo: Analyzers
# DeleteAnalyzerOperation
# PutAnalyzersOperation

# todo: Changes
# IndexChange
# DatabaseChangesOptions
# DocumentChange
# TimeSeriesChange
# CounterChange
# IDatabaseChanges
# DatabaseChange
# OperationStatusChange
# IDatabaseChanges
# DatabaseChanges
# IConnectableChanges
# IChangesObservable
# ChangesObservable
# DatabaseConnectionState
# IChangesConnectionState

# todo: Smuggler
# DatabaseItemType
# DatabaseRecordItemType
# DatabaseSmuggler
# DatabaseSmugglerExportOptions
# IDatabaseSmugglerExportOptions
# DatabaseSmugglerImportOptions
# IDatabaseSmugglerImportOptions
# DatabaseSmugglerOptions
# IDatabaseSmugglerOptions

# todo: Certificates
# AddDatabaseNodeOperation
# PromoteDatabaseNodeOperation
# DeleteServerWideAnalyzerOperation
# PutServerWideAnalyzersOperation
# DocumentCompressionConfigurationResult
# UpdateDocumentsCompressionConfigurationOperation
# IServerWideTask
# DeleteServerWideTaskOperation
# SetDatabasesLockOperation
# ToggleServerWideTaskStateOperation
# GetServerWideExternalReplicationOperation
# PutServerWideExternalReplicationOperation
# ServerWideTaskResponse
# ServerWideExternalReplication
# DeleteServerWideSorterOperation
# PutServerWideSortersOperation



from __future__ import annotations

import datetime
from abc import abstractmethod
from enum import Enum
from typing import Dict
from typing import TYPE_CHECKING
from ravendb.http.topology import ClusterTopology, NodeStatus

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class ClusterTopologyResponse:
    def __init__(self, leader: str, node_tag: str, topology: ClusterTopology, etag: int, status: Dict[str, NodeStatus]):
        self.leader = leader
        self.node_tag = node_tag
        self.topology = topology
        self.status = status
        self.etag = etag

    @classmethod
    def from_json(cls, json_dict: dict) -> ClusterTopologyResponse:
        return cls(
            json_dict["Leader"],
            json_dict["NodeTag"],
            ClusterTopology.from_json(json_dict["Topology"]),
            json_dict["Etag"],
            json_dict["Status"],
        )


class ReadBalanceBehavior(Enum):
    NONE = "None"
    ROUND_ROBIN = "RoundRobin"
    FASTEST_NODE = "FastestNode"

    def __str__(self):
        return self.value


class LoadBalanceBehavior(Enum):
    NONE = "None"
    USE_SESSION_CONTEXT = "UseSessionContext"

    def __str__(self):
        return self.value


class ResponseDisposeHandling(Enum):
    MANUALLY = "Manually"
    AUTOMATIC = "Automatic"

    def __str__(self):
        return self.value


class AggressiveCacheMode(Enum):
    TRACK_CHANGES = "TrackChanges"
    DO_NOT_TRACK_CHANGES = "DoNotTrackChanges"


class AggressiveCacheOptions:
    def __init__(self, duration: datetime.timedelta, mode: AggressiveCacheMode):
        self.duration = duration
        self.mode = mode


class Broadcast:
    @abstractmethod
    def prepare_to_broadcast(self, conventions: DocumentConventions):
        pass



from abc import abstractmethod
from typing import TypeVar, Union, Dict, Callable, Optional, Generic, TYPE_CHECKING

from ravendb.documents.commands.query import QueryCommand
from ravendb.documents.queries.facets.builders import FacetBuilder
from ravendb.documents.queries.facets.definitions import FacetBase
from ravendb.documents.queries.facets.misc import FacetResult
from ravendb.documents.queries.index_query import IndexQuery
from ravendb.documents.queries.query import QueryResult
from ravendb.documents.session.operations.lazy import LazyAggregationQueryOperation
from ravendb.documents.session.operations.query import QueryOperation
from ravendb.documents.store.lazy import Lazy
from ravendb.tools.utils import Stopwatch

_T = TypeVar("_T")

if TYPE_CHECKING:
    from ravendb.documents.session import InMemoryDocumentSessionOperations
    from ravendb.documents.session.document_session import DocumentSession
    from ravendb.documents.session.query import DocumentQuery, RawDocumentQuery


class AggregationQueryBase:
    def __init__(self, session: "InMemoryDocumentSessionOperations"):
        self.__session = session
        self.__query: Union[None, IndexQuery] = None
        self.__duration: Union[None, Stopwatch] = None

    def execute(self) -> Dict[str, FacetResult]:
        command = self.__command
        self.__duration = Stopwatch.create_started()

        self.__session.increment_requests_count()
        self.__session.advanced.request_executor.execute_command(command)

        return self.__process_results(command.result)

    def execute_lazy(
        self, on_eval: Optional[Callable[[Dict[str, FacetResult]], None]] = None
    ) -> Lazy[Dict[str, FacetResult]]:
        self.__query = self._get_index_query()
        self.__session: "DocumentSession"
        return self.__session.add_lazy_operation(
            dict,
            LazyAggregationQueryOperation(
                self.__session,
                self.__query,
                lambda result: self._invoke_after_query_executed(result),
                self.__process_results,
            ),
            on_eval,
        )

    @property
    def __command(self) -> QueryCommand:
        self.__query = self._get_index_query()
        return QueryCommand(self.__session, self.__query, False, False)

    @abstractmethod
    def _get_index_query(self, update_after_query_executed: Optional[bool] = None) -> IndexQuery:
        pass

    @abstractmethod
    def _invoke_after_query_executed(self, result: QueryResult) -> None:
        pass

    def __process_results(self, query_results: QueryResult) -> Dict[str, FacetResult]:
        self._invoke_after_query_executed(query_results)

        results: Dict[str, FacetResult] = {}

        for result in query_results.results:
            facet_result = FacetResult.from_json(result)
            results[facet_result.name] = facet_result

        self.__session.register_includes(query_results.includes)

        QueryOperation.ensure_is_acceptable(
            query_results, self.__query.wait_for_non_stale_results, self.__duration, self.__session
        )
        return results

    def to_string(self):
        return str(self._get_index_query(False))


class AggregationDocumentQuery(Generic[_T], AggregationQueryBase):
    def __init__(self, source: "DocumentQuery[_T]"):
        super().__init__(source.session)
        self.__source = source

    def and_aggregate_by(self, builder_or_facet: Union[Callable[[FacetBuilder[_T]], None], FacetBase]):
        if isinstance(builder_or_facet, FacetBase):
            facet = builder_or_facet
        else:
            f = FacetBuilder()
            builder_or_facet(f)
            facet = f.facet

        self.__source.aggregate_by(facet)
        return self

    def _get_index_query(self, update_after_query_executed: Optional[bool] = True) -> IndexQuery:
        return self.__source.index_query

    def _invoke_after_query_executed(self, result: QueryResult) -> None:
        self.__source.invoke_after_query_executed(result)


class AggregationRawDocumentQuery(Generic[_T], AggregationQueryBase):
    def __init__(self, source: "RawDocumentQuery[_T]", session: "InMemoryDocumentSessionOperations"):
        super().__init__(session)
        self.__source = source

        if source is None:
            raise ValueError("Source cannot be None")

    def _get_index_query(self, update_after_query_executed: Optional[bool] = True) -> IndexQuery:
        return self.__source.index_query

    def _invoke_after_query_executed(self, result: QueryResult) -> None:
        self.__source.invoke_after_query_executed(result)



from __future__ import annotations
from enum import Enum
from typing import Optional, Type, TypeVar

from ravendb.documents.indexes.definitions import AbstractCommonApiForIndexes


class SearchOperator(Enum):
    OR = "Or"
    AND = "And"


_TIndex = TypeVar("_TIndex", bound=AbstractCommonApiForIndexes)


class Query:
    def __init__(self):
        self.__collection_name: Optional[str] = None
        self.__index_name: Optional[str] = None
        self.__index_type: Optional[Type[_TIndex]] = None

    @property
    def collection(self) -> str:
        return self.__collection_name

    @property
    def index_name(self) -> str:
        return self.__index_name

    @property
    def index_type(self) -> Type[_TIndex]:
        return self.__index_type

    @classmethod
    def from_collection_name(cls, collection_name: str) -> Query:
        query = cls()
        query.__collection_name = collection_name
        return query

    @classmethod
    def from_index_name(cls, index_name: str) -> Query:
        query = cls()
        query.__index_name = index_name
        return query

    @classmethod
    def from_index_type(cls, index_type: Type[_TIndex]) -> Query:
        query = cls()
        query.__index_type = index_type
        return query



import json
from typing import TYPE_CHECKING, Set

import requests

from ravendb.http.topology import RaftCommand
from ravendb.documents.operations.definitions import VoidMaintenanceOperation
from ravendb.documents.smuggler.common import DatabaseItemType
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.serverwide.operations.common import ServerOperation
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class AdminJsConsoleOperation(ServerOperation[dict]):
    def __init__(self, script: str):
        self.__script = script

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[dict]:
        return AdminJsConsoleOperation.__AdminJsConsoleCommand(self.__script)

    class __AdminJsConsoleCommand(RavenCommand[dict]):
        def __init__(self, script: str):
            super().__init__(dict)
            self.__script = script

        def create_request(self, node: ServerNode) -> requests.Request:
            request = requests.Request("POST", f"{node.url}/admin/console?serverScript=true")
            request.data = {"Script": self.__script}
            return request

        def is_read_request(self) -> bool:
            return False

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = json.loads(response)


class CreateSampleDataOperation(VoidMaintenanceOperation):
    def __init__(self, operate_on_types=None):
        if operate_on_types is None:
            operate_on_types = {DatabaseItemType.DOCUMENTS}
        self._operate_on_types = operate_on_types

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.CreateSampleDataCommand(self._operate_on_types)

    class CreateSampleDataCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, operate_on_types: Set[DatabaseItemType]):
            super().__init__()
            self._operate_on_types = operate_on_types

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/studio/sample-data"
            url += f"?{'&'.join([f'operateOnTypes={x.value}' for x in self._operate_on_types])}"
            return requests.Request("POST", url)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



from __future__ import annotations

import base64
import enum
import json
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List

import requests

from ravendb.http.raven_command import RavenCommand, RavenCommandResponseType, VoidRavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import RaftCommand
from ravendb.serverwide.operations.common import ServerOperation, VoidServerOperation
from ravendb.tools.utils import CaseInsensitiveDict, Utils
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class SecurityClearance(enum.Enum):
    UNAUTHENTICATED_CLIENTS = "UnauthenticatedClients"
    CLUSTER_ADMIN = "ClusterAdmin"
    CLUSTER_NODE = "ClusterNode"
    OPERATOR = "Operator"
    VALID_USER = "ValidUser"

    def __str__(self):
        return self.value


class DatabaseAccess(enum.Enum):
    READ_WRITE = "ReadWrite"
    ADMIN = "Admin"
    READ = "Read"

    def __str__(self):
        return self.value


class CertificateRawData:
    def __init__(self, raw_data: bytes = None):
        self.raw_data = raw_data


class CertificateMetadata:
    def __init__(
        self,
        name: str = None,
        security_clearance: SecurityClearance = None,
        thumbprint: str = None,
        not_after: datetime = None,
        permissions: CaseInsensitiveDict[str, DatabaseAccess] = None,
        collection_secondary_keys: List[str] = None,
        collection_primary_key: str = None,
        public_key_pinning_hash: str = None,
        not_before: datetime = None,
    ):
        self.name = name
        self.security_clearance = security_clearance
        self.thumbprint = thumbprint
        self.not_after = not_after
        self.permissions = permissions
        self.collection_primary_key = collection_primary_key
        self.collection_secondary_keys = collection_secondary_keys
        self.public_key_pinning_hash = public_key_pinning_hash
        self.not_before = not_before

    @classmethod
    def from_json(cls, json_dict: dict) -> CertificateMetadata:
        return cls(
            json_dict["Name"],
            SecurityClearance(json_dict.get("SecurityClearance", None)),
            json_dict.get("Thumbprint", None),
            Utils.string_to_datetime(json_dict["NotAfter"]) if "NotAfter" in json_dict else None,
            json_dict.get("Permissions", None),
            json_dict.get("CollectionSecondaryKeys", None),
            json_dict.get("CollectionPrimaryKey", None),
            json_dict.get("PublicKeyPinningHash", None),
            Utils.string_to_datetime(json_dict["NotBefore"]) if "NotBefore" in json_dict else None,
        )


class CertificateDefinition(CertificateMetadata):
    def __init__(
        self,
        certificate: str = None,
        password: str = None,
        name: str = None,
        security_clearance: SecurityClearance = None,
        thumbprint: str = None,
        not_after: datetime = None,
        permissions: CaseInsensitiveDict[str, DatabaseAccess] = None,
        collection_secondary_keys: List[str] = None,
        collection_primary_key: str = None,
        public_key_pinning_hash: str = None,
    ):
        super().__init__(
            name,
            security_clearance,
            thumbprint,
            not_after,
            permissions,
            collection_secondary_keys,
            collection_primary_key,
            public_key_pinning_hash,
        )
        self.certificate = certificate
        self.password = password

    def to_json(self) -> dict:
        json_dict = {
            "Name": self.name,
            "SecurityClearance": self.security_clearance,
            "Thumbprint": self.thumbprint,
            "Permissions": self.permissions,
            "CollectionSecondaryKeys": self.collection_secondary_keys,
            "CollectionPrimaryKey": self.collection_primary_key,
            "PublicKeyPinningHash": self.public_key_pinning_hash,
            "Certificate": self.certificate,
            "Password": self.password,
        }
        if self.not_after:
            json_dict.update({"NotAfter": Utils.datetime_to_string(self.not_after)})
        return json_dict

    @classmethod
    def from_json(cls, json_dict: dict) -> CertificateDefinition:
        return cls(
            json_dict["Certificate"],
            json_dict.get("Password", None),
            json_dict["Name"],
            json_dict["SecurityClearance"],
            json_dict["Thumbprint"],
            json_dict["NotAfter"],
            {item[0]: DatabaseAccess(item[1]) for item in json_dict["Permissions"].items()},
            json_dict["CollectionSecondaryKeys"],
            json_dict["CollectionPrimaryKey"],
            json_dict["PublicKeyPinningHash"],
        )


class CreateClientCertificateOperation(ServerOperation[CertificateRawData]):
    def __init__(
        self, name: str, permissions: Dict[str, DatabaseAccess], clearance: SecurityClearance, password: str = None
    ):
        if name is None:
            raise ValueError("name cannot be None")
        if permissions is None:
            raise ValueError("permissions cannot be None")

        super(CreateClientCertificateOperation, self).__init__()
        self.__name = name
        self.__permissions = permissions
        self.__clearance = clearance
        self.__password = password

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[CertificateRawData]:
        return self.__CreateClientCertificateCommand(self.__name, self.__permissions, self.__clearance, self.__password)

    class __CreateClientCertificateCommand(RavenCommand[CertificateRawData], RaftCommand):
        def __init__(self, name, permissions, clearance, password):
            if name is None:
                raise ValueError("name cannot be None")
            if permissions is None:
                raise ValueError("permissions cannot be None")

            super().__init__(CertificateRawData)

            self.__name = name
            self.__permissions = permissions
            self.__clearance = clearance
            self.__password = password

            self._response_type = RavenCommandResponseType.RAW

        def is_read_request(self) -> bool:
            return False

        def create_request(self, server_node):
            url = server_node.url + "/admin/certificates"
            data = {"Name": self.__name, "SecurityClearance": str(self.__clearance)}
            if self.__password:
                data["Password"] = self.__password

            permissions = {}
            for key, value in self.__permissions.items():
                permissions.update({key: str(value)})

            data["Permissions"] = permissions

            return requests.Request("POST", url, data=data)

        def set_response_raw(self, response: requests.Response, stream: bytes) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = CertificateRawData()

            try:
                self.result.raw_data = bytearray(stream)
            except Exception as e:
                self._throw_invalid_response(e)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class GetCertificatesResponse:
    def __init__(self, results: List[CertificateDefinition]):
        self.results = results

    @classmethod
    def from_json(cls, json_dict: dict) -> GetCertificatesResponse:
        results = []
        for cert_dict in json_dict["Results"]:
            results.append(CertificateDefinition.from_json(cert_dict))
        return cls(results)


class GetCertificateOperation(ServerOperation[CertificateDefinition]):
    def __init__(self, thumbprint: str):
        if thumbprint is None:
            raise ValueError("thumbprint cannot be None")

        self.__thumbprint = thumbprint

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[CertificateDefinition]:
        return self.__GetCertificateCommand(self.__thumbprint)

    class __GetCertificateCommand(RavenCommand[CertificateDefinition]):
        def __init__(self, thumbprint: str):
            if thumbprint is None:
                raise ValueError("thumbprint cannot be None")

            super().__init__(CertificateDefinition)
            self.__thumbprint = thumbprint

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/certificates?thumbprint={Utils.escape(self.__thumbprint, False,False)}"
            return requests.Request("GET", url)

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return

            certificates = GetCertificatesResponse.from_json(json.loads(response))
            if len(certificates.results) != 1:
                self._throw_invalid_response()

            self.result = certificates.results[0]


class GetCertificatesOperation(ServerOperation[List[CertificateDefinition]]):
    def __init__(self, start: int, page_size: int):
        self.__start = start
        self.__page_size = page_size

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[CertificateDefinition]:
        return self.__GetCertificatesCommand(self.__start, self.__page_size)

    class __GetCertificatesCommand(RavenCommand[List[CertificateDefinition]]):
        def __init__(self, start: int, page_size: int):
            super().__init__(list)
            self.__start = start
            self.__page_size = page_size

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/certificates?start={self.__start}&pageSize={self.__page_size}"
            return requests.Request("GET", url)

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return

            certificates = GetCertificatesResponse.from_json(json.loads(response))
            self.result = certificates.results


class GetCertificateMetadataOperation(ServerOperation[CertificateMetadata]):
    def __init__(self, thumbprint: str):
        if thumbprint is None:
            raise ValueError("thumbprint cannot be None")
        self.__thumbprint = thumbprint

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[CertificateMetadata]:
        return self.__GetCertificateMetadataCommand(self.__thumbprint)

    class __GetCertificateMetadataCommand(RavenCommand[CertificateMetadata]):
        def __init__(self, thumbprint: str):
            super().__init__(CertificateMetadata)
            self.__thumbprint = thumbprint

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            path = (
                f"{node.url}/admin/certificates"
                f"?thumbprint={Utils.escape(self.__thumbprint,True, False)}"
                f"&metadataOnly=true"
            )

            return requests.Request("GET", path)

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return

            results = json.loads(response)["Results"]
            results = list(map(CertificateMetadata.from_json, results))

            if len(results) != 1:
                self._throw_invalid_response()

            self.result = results[0]


class GetCertificatesMetadataOperation(ServerOperation[List[CertificateMetadata]]):
    def __init__(self, name: str = None):
        self.__name = name

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[List[CertificateMetadata]]:
        return self.__GetCertificatesMetadataCommand(self.__name)

    class __GetCertificatesMetadataCommand(RavenCommand[List[CertificateMetadata]]):
        def __init__(self, name: str):
            super().__init__(list)
            self.__name = name

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            path = [node.url, "/admin/certificates?metadataOnly=true"]

            if self.__name:
                path.append("&name=")
                path.append(Utils.quote_key(self.__name))

            url = "".join(path)

            return requests.Request("GET", url)

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return
            self.result = []
            for metadata in json.loads(response)["Results"]:
                self.result.append(CertificateMetadata.from_json(metadata))


class DeleteCertificateOperation(VoidServerOperation):
    def __init__(self, thumbprint: str):
        if thumbprint is None:
            raise ValueError("thumbprint cannot be None")

        super(DeleteCertificateOperation, self).__init__()
        self.__thumbprint = thumbprint

    def get_command(self, conventions: "DocumentConventions") -> VoidRavenCommand:
        return self.__DeleteCertificateCommand(self.__thumbprint)

    class __DeleteCertificateCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, thumbprint: str):
            if thumbprint is None:
                raise ValueError("certificate cannot be None")
            super().__init__()
            self.__thumbprint = thumbprint

        def create_request(self, server_node: ServerNode) -> requests.Request:
            url = f"{server_node.url}/admin/certificates?thumbprint={self.__thumbprint}"
            return requests.Request("DELETE", url)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class PutClientCertificateOperation(VoidServerOperation):
    def __init__(
        self, name: str, certificate: str, permissions: Dict[str, DatabaseAccess], clearance: SecurityClearance
    ):
        if certificate is None:
            raise ValueError("certificate cannot be None")

        if permissions is None:
            raise ValueError("permissions cannot be None")

        if name is None:
            raise ValueError("name cannot be None")

        self.__certificate = certificate
        self.__permissions = permissions
        self.__name = name
        self.__clearance = clearance

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.__PutClientCertificateCommand(self.__name, self.__certificate, self.__permissions, self.__clearance)

    class __PutClientCertificateCommand(VoidRavenCommand, RaftCommand):
        def __init__(
            self, name: str, certificate: str, permissions: Dict[str, DatabaseAccess], clearance: SecurityClearance
        ):
            if certificate is None:
                raise ValueError("certificate cannot be None")

            if permissions is None:
                raise ValueError("permissions cannot be None")

            super().__init__()
            self.__certificate = certificate
            self.__permissions = permissions
            self.__name = name
            self.__clearance = clearance

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/certificates"
            request = requests.Request("PUT", url)

            request.data = {
                "Name": self.__name,
                "Certificate": self.__certificate,
                "SecurityClearance": str(self.__clearance),
                "Permissions": self.__permissions,
            }

            return request

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class EditClientCertificateOperation(VoidServerOperation):
    class Parameters:
        def __init__(
            self, thumbprint: str, permissions: Dict[str, DatabaseAccess], name: str, clearance: SecurityClearance
        ):
            self.thumbprint = thumbprint
            self.permissions = permissions
            self.name = name
            self.clearance = clearance

    def __init__(self, parameters: Parameters):
        if parameters is None:
            raise ValueError("parameters cannot be None")

        if parameters.name is None:
            raise ValueError("name cannot be None")

        if parameters.thumbprint is None:
            raise ValueError("thumbprint cannot be None")

        if parameters.permissions is None:
            raise ValueError("permissions cannot be None")

        self.__name = parameters.name
        self.__thumbprint = parameters.thumbprint
        self.__permissions = parameters.permissions
        self.__clearance = parameters.clearance

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.__EditCertificateClientCommand(self.__thumbprint, self.__name, self.__permissions, self.__clearance)

    class __EditCertificateClientCommand(VoidRavenCommand, RaftCommand):
        def __init__(
            self, thumbprint: str, name: str, permissions: Dict[str, DatabaseAccess], clearance: SecurityClearance
        ):
            super().__init__()
            self.__thumbprint = thumbprint
            self.__name = name
            self.__permissions = permissions
            self.__clearance = clearance

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/certificates/edit"

            definition = CertificateDefinition()
            definition.thumbprint = self.__thumbprint
            definition.permissions = self.__permissions
            definition.security_clearance = self.__clearance
            definition.name = self.__name

            request = requests.Request("POST", url)
            request.data = definition.to_json()

            return request

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class ReplaceClusterCertificateOperation(VoidServerOperation):
    def __init__(self, cert_bytes: bytes, replace_immediately: bool):
        if cert_bytes is None:
            raise ValueError("cert_bytes cannot be None")

        self.__cert_bytes = cert_bytes
        self.__replace_immediately = replace_immediately

    @property
    def cert_bytes(self) -> bytes:
        return self.__cert_bytes

    @property
    def replace_immediately(self) -> bool:
        return self.__replace_immediately

    def get_command(self, conventions: "DocumentConventions") -> VoidRavenCommand:
        return self.__ReplaceClusterCertificateCommand(self.__cert_bytes, self.__replace_immediately)

    class __ReplaceClusterCertificateCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, cert_bytes: bytes, replace_immediately: bool):
            super().__init__()
            if cert_bytes is None:
                raise ValueError("cert_bytes cannot be None")

            self.__cert_bytes = cert_bytes
            self.__replace_immediately = replace_immediately

        @property
        def cert_bytes(self) -> bytes:
            return self.__cert_bytes

        @property
        def replace_immediately(self) -> bool:
            return self.__replace_immediately

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = (
                f"{node.url}/admin/certificates/replace-cluster-cert"
                f"?replaceImmediately={'true' if self.__replace_immediately else 'false'}"
            )
            request = requests.Request("POST", url)
            request.data = json.loads(json.dumps({"Certificate": base64.b64encode(self.__cert_bytes).decode("utf-8")}))
            return request

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



import datetime
import json
import uuid
from typing import Optional

import requests

from ravendb.documents.commands.subscriptions import TcpConnectionInfo
from ravendb.http.misc import ClusterTopologyResponse
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import Topology
from ravendb.tools.utils import Utils


class GetDatabaseTopologyCommand(RavenCommand):
    def __init__(self, debug_tag: Optional[str] = None, application_identifier: Optional[uuid.UUID] = None):
        super().__init__(Topology)
        self.__debug_tag = debug_tag
        self.__application_identifier = application_identifier

    def is_read_request(self) -> bool:
        return True

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/topology?name={node.database}{('&'+self.__debug_tag) if self.__debug_tag else ''}"
        if self.__application_identifier:
            url += f"&applicationIdentifier=" + str(self.__application_identifier)
        if ".fiddler" in node.url.lower():
            url += f"&localUrl={Utils.escape(node.url,False,False)}"
        return requests.Request(method="GET", url=url)

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is None:
            return

        # todo: that's pretty bad way to do that, replace with initialization function that take nested object types
        self.result: Topology = Utils.initialize_object(json.loads(response), self._result_class, True)
        node_list = []
        for node in self.result.nodes:
            node_list.append(Utils.initialize_object(node, ServerNode, True))
        self.result.nodes = node_list


class GetClusterTopologyCommand(RavenCommand[ClusterTopologyResponse]):
    def __init__(self, debug_tag: str = None):
        super(GetClusterTopologyCommand, self).__init__(ClusterTopologyResponse)
        self.__debug_tag = debug_tag

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/cluster/topology"
        if self.__debug_tag is not None:
            url += f"?{self.__debug_tag}"

        return requests.Request("GET", url)

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is None:
            super()._throw_invalid_response()

        self.result: ClusterTopologyResponse = ClusterTopologyResponse.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return True


class GetTcpInfoCommand(RavenCommand[TcpConnectionInfo]):
    def __init__(self, tag: str, db_name: Optional[str] = None):
        super(GetTcpInfoCommand, self).__init__(TcpConnectionInfo)
        self._tag = tag
        self._db_name = db_name
        self._timeout = datetime.timedelta(seconds=15)
        self._requested_node: Optional[ServerNode] = None

    def create_request(self, node: ServerNode) -> requests.Request:
        url = (
            f"{node.url}/info/tcp?tcp={self._tag}"
            if not self._db_name
            else f"{node.url}/databases/{self._db_name}/info/tcp?tag={self._tag}"
        )
        self._requested_node = node
        return requests.Request("GET", url)

    @property
    def requested_node(self) -> ServerNode:
        return self._requested_node

    @requested_node.setter
    def requested_node(self, value: ServerNode):
        self._requested_node = value

    def set_response(self, response: Optional[str], from_cache: bool) -> None:
        if response is None:
            self._throw_invalid_response()

        self.result = TcpConnectionInfo.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return True



from typing import List, TYPE_CHECKING

import requests

from ravendb import ServerNode
from ravendb.http.raven_command import VoidRavenCommand
from ravendb.http.topology import RaftCommand
from ravendb.documents.queries.sorting import SorterDefinition
from ravendb.serverwide.operations.common import VoidServerOperation
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class PutServerWideSortersOperation(VoidServerOperation):
    def __init__(self, *sorters_to_add: SorterDefinition):
        if not sorters_to_add:
            raise ValueError("Sorters must not be empty or None")

        self._sorters_to_add = sorters_to_add

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.PutServerWideSortersCommand(conventions, list(self._sorters_to_add))

    class PutServerWideSortersCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, conventions: "DocumentConventions", sorters_to_add: List[SorterDefinition]):
            super().__init__()
            if conventions is None:
                raise ValueError("Conventions cannot be None")
            if sorters_to_add is None:
                raise ValueError("Sorters cannot be None")

            self._sorters_to_add = []

            for sorter in sorters_to_add:
                if sorter.name is None:
                    raise ValueError("Sorter name cannot be None")

                self._sorters_to_add.append(sorter.to_json())

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/sorters"

            return requests.Request(
                "PUT",
                url,
                data={
                    "Sorters": self._sorters_to_add,
                },
            )

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class DeleteServerWideSorterOperation(VoidServerOperation):
    def __init__(self, sorter_name: str):
        if sorter_name is None:
            raise ValueError("SorterName cannot be None")

        self._sorter_name = sorter_name

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.DeleteServerWideSorterCommand(self._sorter_name)

    class DeleteServerWideSorterCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, sorter_name: str) -> "VoidRavenCommand":
            super().__init__()
            if sorter_name is None:
                raise ValueError("Sorter name cannot be None")

            self._sorter_name = sorter_name

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("DELETE", f"{node.url}/admin/sorters?name={Utils.quote_key(self._sorter_name)}")

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



# todo: implement
from datetime import datetime
from typing import Optional, List, Dict

from ravendb.tools.utils import Utils


class ReplicationNode:
    pass


class ExternalReplicationBase(ReplicationNode):
    pass


class ExternalReplication(ExternalReplicationBase):
    pass


class PullReplicationAsSink(ExternalReplicationBase):
    pass


class PullReplicationDefinition:
    pass


class DetailedReplicationHubAccess:
    def __init__(
        self,
        name: Optional[str] = None,
        thumbprint: Optional[str] = None,
        certificate: Optional[str] = None,
        not_before: Optional[datetime] = None,
        not_after: Optional[datetime] = None,
        subject: Optional[str] = None,
        issuer: Optional[str] = None,
        allowed_hub_to_sink_paths: Optional[List[str]] = None,
        allowed_sink_to_hub_paths: Optional[List[str]] = None,
    ):
        self.name = name
        self.thumbprint = thumbprint
        self.certificate = certificate
        self.not_before = not_before
        self.not_after = not_after
        self.subject = subject
        self.issuer = issuer
        self.allowed_hub_to_sink_paths = allowed_hub_to_sink_paths
        self.allowed_sink_to_hub_paths = allowed_sink_to_hub_paths

    def to_json(self) -> Dict:
        return {
            "Name": self.name,
            "Thumbprint": self.thumbprint,
            "Certificate": self.certificate,
            "NotBefore": Utils.datetime_to_string(self.not_before),
            "NotAfter": Utils.datetime_to_string(self.not_after),
            "Subject": self.subject,
            "Issuer": self.issuer,
            "AllowedHubToSinkPaths": self.allowed_hub_to_sink_paths,
            "AllowedSinkToHubPaths": self.allowed_sink_to_hub_paths,
        }



from __future__ import annotations
from abc import ABC
from datetime import datetime, timedelta
from enum import Enum
from typing import Dict, Any

from ravendb.tools.utils import Utils


class BackupType(Enum):
    BACKUP = "Backup"
    SNAPSHOT = "Snapshot"


class CompressionLevel(Enum):
    OPTIMAL = "Optimal"
    FASTEST = "Fastest"
    NO_COMPRESSION = "NoCompression"


class EncryptionMode(Enum):
    NONE = "None"
    USE_DATABASE_KEY = "UseDatabaseKey"
    USE_PROVIDED_KEY = "UseProvidedKey"


class GetBackupConfigurationScript:
    def __init__(self, exec: str = None, arguments: str = None, timeout_in_ms: int = None):
        self.exec = exec
        self.arguments = arguments
        self.timeout_in_ms = timeout_in_ms

    @classmethod
    def default(cls) -> GetBackupConfigurationScript:
        return cls(timeout_in_ms=10000)

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> GetBackupConfigurationScript:
        return cls(json_dict["Exec"], json_dict["Arguments"], json_dict["TimeoutInMs"])

    def to_json(self) -> Dict[str, Any]:
        return {"Exec": self.exec, "Arguments": self.arguments, "TimeoutInMs": self.timeout_in_ms}


class BackupSettings(ABC):
    def __init__(self, disabled: bool = None, get_backup_configuration_script: GetBackupConfigurationScript = None):
        self.disabled = disabled
        self.get_backup_configuration_script = get_backup_configuration_script

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "GetBackupConfigurationScript": self.get_backup_configuration_script.to_json(),
        }


class LocalSettings(BackupSettings):
    def __init__(
        self,
        disabled: bool = None,
        get_backup_configuration_script: GetBackupConfigurationScript = None,
        folder_path: str = None,
    ):
        super().__init__(disabled, get_backup_configuration_script)
        self.folder_path = folder_path

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> LocalSettings:
        return cls(
            json_dict["Disabled"],
            GetBackupConfigurationScript.from_json(json_dict["GetBackupConfigurationScript"]),
            json_dict["FolderPath"],
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "GetBackupConfigurationScript": self.get_backup_configuration_script.to_json(),
            "FolderPath": self.folder_path,
        }


class AmazonSettings(BackupSettings, ABC):
    def __init__(
        self,
        disabled: bool = None,
        get_backup_configuration_script: GetBackupConfigurationScript = None,
        aws_access_key: str = None,
        aws_secret_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None,
        remote_folder_name: str = None,
    ):
        super().__init__(disabled, get_backup_configuration_script)
        self.aws_access_key = aws_access_key
        self.aws_secret_key = aws_secret_key
        self.aws_session_token = aws_session_token
        self.aws_region_name = aws_region_name
        self.remote_folder_name = remote_folder_name


class S3Settings(AmazonSettings):
    def __init__(
        self,
        disabled: bool = None,
        get_backup_configuration_script: GetBackupConfigurationScript = None,
        aws_access_key: str = None,
        aws_secret_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None,
        remote_folder_name: str = None,
        bucket_name: str = None,
        custom_server_url: str = None,
        force_path_style: bool = None,
    ):
        super().__init__(
            disabled,
            get_backup_configuration_script,
            aws_access_key,
            aws_secret_key,
            aws_session_token,
            aws_region_name,
            remote_folder_name,
        )
        self.bucket_name = bucket_name
        self.custom_server_url = custom_server_url
        self.force_path_style = force_path_style

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> S3Settings:
        return cls(
            json_dict["Disabled"],
            GetBackupConfigurationScript.from_json(json_dict["GetBackupConfigurationScript"]),
            json_dict["AwsAccessKey"],
            json_dict["AwsSecretKey"],
            json_dict["AwsSessionToken"],
            json_dict["AwsRegionName"],
            json_dict["RemoteFolderName"],
            json_dict["BucketName"],
            json_dict["CustomServerUrl"],
            json_dict["ForcePathStyle"],
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "GetBackupConfigurationScript": self.get_backup_configuration_script.to_json(),
            "AwsAccessKey": self.aws_access_key,
            "AwsSecretKey": self.aws_secret_key,
            "AwsSessionToken": self.aws_session_token,
            "AwsRegionName": self.aws_region_name,
            "RemoteFolderName": self.remote_folder_name,
            "BucketName": self.bucket_name,
            "CustomServerUrl": self.custom_server_url,
            "ForcePathStyle": self.force_path_style,
        }


class GlacierSettings(AmazonSettings):
    def __init__(
        self,
        disabled: bool = None,
        get_backup_configuration_script: GetBackupConfigurationScript = None,
        aws_access_key: str = None,
        aws_secret_key: str = None,
        aws_session_token: str = None,
        aws_region_name: str = None,
        remote_folder_name: str = None,
        vault_name: str = None,
    ):
        super().__init__(
            disabled,
            get_backup_configuration_script,
            aws_access_key,
            aws_secret_key,
            aws_session_token,
            aws_region_name,
            remote_folder_name,
        )
        self.vault_name = vault_name

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> GlacierSettings:
        return cls(
            json_dict["Disabled"],
            GetBackupConfigurationScript.from_json(json_dict["GetBackupConfigurationScript"]),
            json_dict["AwsAccessKey"],
            json_dict["AwsSecretKey"],
            json_dict["AwsSessionToken"],
            json_dict["AwsRegionName"],
            json_dict["RemoteFolderName"],
            json_dict["VaultName"],
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "GetBackupConfigurationScript": self.get_backup_configuration_script.to_json(),
            "AwsAccessKey": self.aws_access_key,
            "AwsSecretKey": self.aws_secret_key,
            "AwsSessionToken": self.aws_session_token,
            "AwsRegionName": self.aws_region_name,
            "RemoteFolderName": self.remote_folder_name,
            "VaultName": self.vault_name,
        }


class AzureSettings(BackupSettings):
    def __init__(
        self,
        disabled: bool = None,
        get_backup_configuration_script: GetBackupConfigurationScript = None,
        storage_container: str = None,
        remote_folder_name: str = None,
        account_name: str = None,
        account_key: str = None,
        sas_token: str = None,
    ):
        super().__init__(disabled, get_backup_configuration_script)
        self.storage_container = storage_container
        self.remote_folder_name = remote_folder_name
        self.account_name = account_name
        self.account_key = account_key
        self.sas_token = sas_token

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> AzureSettings:
        return cls(
            json_dict["Disabled"],
            (
                GetBackupConfigurationScript.from_json(json_dict["GetBackupConfigurationScript"])
                if json_dict["GetBackupConfigurationScript"]
                else None
            ),
            json_dict["StorageContainer"],
            json_dict["RemoteFolderName"],
            json_dict["AccountName"],
            json_dict["AccountKey"],
            json_dict["SasToken"],
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "GetBackupConfigurationScript": (
                self.get_backup_configuration_script.to_json() if self.get_backup_configuration_script else None
            ),
            "StorageContainer": self.storage_container,
            "RemoteFolderName": self.remote_folder_name,
            "AccountName": self.account_name,
            "AccountKey": self.account_key,
            "SasToken": self.sas_token,
        }


class FtpSettings(BackupSettings):
    def __init__(
        self,
        disabled: bool = None,
        get_backup_configuration_script: GetBackupConfigurationScript = None,
        url: str = None,
        port: int = None,
        user_name: str = None,
        password: str = None,
        certificate_as_base64: str = None,
        certificate_file_name: str = None,
    ):
        super().__init__(disabled, get_backup_configuration_script)
        self.url = url
        self.port = port
        self.user_name = user_name
        self.password = password
        self.certificate_as_base64 = certificate_as_base64
        self.certificate_file_name = certificate_file_name

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> FtpSettings:
        return cls(
            json_dict["Disabled"],
            (
                GetBackupConfigurationScript.from_json(json_dict["GetBackupConfigurationScript"])
                if json_dict["GetBackupConfigurationScript"]
                else None
            ),
            json_dict["Url"],
            json_dict["Port"] if "Port" in json_dict else None,
            json_dict["UserName"],
            json_dict["Password"],
            json_dict["CertificateAsBase64"],
            json_dict["CertificateFileName"] if "CertificateFileName" in json_dict else None,
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "GetBackupConfigurationScript": (
                self.get_backup_configuration_script.to_json() if self.get_backup_configuration_script else None
            ),
            "Url": self.url,
            "Port": self.port,
            "UserName": self.user_name,
            "Password": self.password,
            "CertificateAsBase64": self.certificate_as_base64,
            "CertificateFileName": self.certificate_file_name,
        }


class BackupStatus(ABC):
    def __init__(
        self,
        last_full_backup: datetime = None,
        last_incremental_backup: datetime = None,
        full_backup_duration_in_ms: int = None,
        incremental_backup_duration_in_ms: int = None,
        exception: str = None,
    ):
        self.last_full_backup = last_full_backup
        self.last_incremental_backup = last_incremental_backup
        self.full_backup_duration_in_ms = full_backup_duration_in_ms
        self.incremental_backup_duration_in_ms = incremental_backup_duration_in_ms
        self.exception = exception


class GoogleCloudSettings(BackupStatus):
    def __init__(
        self,
        last_full_backup: datetime = None,
        last_incremental_backup: datetime = None,
        full_backup_duration_in_ms: int = None,
        incremental_backup_duration_in_ms: int = None,
        exception: str = None,
        bucket_name: str = None,
        remote_folder_name: str = None,
        google_credentials_json: str = None,
    ):
        super().__init__(
            last_full_backup,
            last_incremental_backup,
            full_backup_duration_in_ms,
            incremental_backup_duration_in_ms,
            exception,
        )
        self.bucket_name = bucket_name
        self.remote_folder_name = remote_folder_name
        self.google_credentials_json = google_credentials_json

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> GoogleCloudSettings:
        return cls(
            Utils.string_to_datetime(json_dict["LastFullBackup"]),
            Utils.string_to_datetime(json_dict["LastIncrementalBackup"]),
            json_dict["FullBackupDurationInMs"],
            json_dict["IncrementalBackupDurationInMs"],
            json_dict["Exception"],
            json_dict["BucketName"],
            json_dict["RemoteFolderName"],
            json_dict["GoogleCredentialsJson"],
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "LastFullBackup": Utils.datetime_to_string(self.last_full_backup),
            "LastIncrementalBackup": Utils.datetime_to_string(self.last_full_backup),
            "FullBackupDurationInMs": self.full_backup_duration_in_ms,
            "IncrementalBackupDurationInMs": self.incremental_backup_duration_in_ms,
            "Exception": self.exception,
            "BucketName": self.bucket_name,
            "RemoteFolderName": self.remote_folder_name,
            "GoogleCredentialsJson": self.google_credentials_json,
        }


class BackupEncryptionSettings:
    def __init__(self, key: str = None, encryption_mode: EncryptionMode = None):
        self.key = key
        self.encryption_mode = encryption_mode

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> BackupEncryptionSettings:
        return BackupEncryptionSettings(json_dict["Key"], EncryptionMode(json_dict["EncryptionMode"]))

    def to_json(self) -> Dict[str, Any]:
        return {"Key": self.key, "EncryptionMode": self.encryption_mode.value}


class SnapshotSettings:
    def __init__(self, compression_level: CompressionLevel = None):
        self.compression_level = compression_level

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> SnapshotSettings:
        return SnapshotSettings(CompressionLevel(json_dict["CompressionLevel"]))

    def to_json(self) -> Dict[str, Any]:
        return {"CompressionLevel": self.compression_level.value}


class BackupConfiguration:
    def __init__(
        self,
        backup_type: BackupType = None,
        snapshot_settings: SnapshotSettings = None,
        backup_encryption_settings: BackupEncryptionSettings = None,
        local_settings: LocalSettings = None,
        s3_settings: S3Settings = None,
        glacier_settings: GlacierSettings = None,
        azure_settings: AzureSettings = None,
        ftp_settings: FtpSettings = None,
        google_cloud_settings: GoogleCloudSettings = None,
    ):
        self.backup_type = backup_type
        self.snapshot_settings = snapshot_settings
        self.backup_encryption_settings = backup_encryption_settings
        self.local_settings = local_settings
        self.s3_settings = s3_settings
        self.glacier_settings = glacier_settings
        self.azure_settings = azure_settings
        self.ftp_settings = ftp_settings
        self.google_cloud_settings = google_cloud_settings

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> BackupConfiguration:
        return cls(
            BackupType(json_dict["BackupType"]),
            SnapshotSettings.from_json(json_dict["SnapshotSettings"]) if json_dict["SnapshotSettings"] else None,
            (
                BackupEncryptionSettings.from_json(json_dict["BackupEncryptionSettings"])
                if json_dict["BackupEncryptionSettings"]
                else None
            ),
            LocalSettings.from_json(json_dict["LocalSettings"]) if json_dict["LocalSettings"] else None,
            S3Settings.from_json(json_dict["S3Settings"]) if json_dict["S3Settings"] else None,
            GlacierSettings.from_json(json_dict["GlacierSettings"]) if json_dict["GlacierSettings"] else None,
            AzureSettings.from_json(json_dict["AzureSettings"]) if json_dict["AzureSettings"] else None,
            FtpSettings.from_json(json_dict["FtpSettings"]) if json_dict["FtpSettings"] else None,
            (
                GoogleCloudSettings.from_json(json_dict["GoogleCloudSettings"])
                if json_dict["GoogleCloudSettings"]
                else None
            ),
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "BackupType": self.backup_type.value if self.backup_type else None,
            "SnapshotSettings": self.snapshot_settings.to_json() if self.snapshot_settings else None,
            "BackupEncryptionSettings": (
                self.backup_encryption_settings.to_json() if self.backup_encryption_settings else None
            ),
            "LocalSettings": self.local_settings.to_json() if self.local_settings else None,
            "S3Settings": self.s3_settings.to_json() if self.s3_settings else None,
            "GlacierSettings": self.glacier_settings.to_json() if self.glacier_settings else None,
            "AzureSettings": self.azure_settings.to_json() if self.azure_settings else None,
            "FtpSettings": self.ftp_settings.to_json() if self.ftp_settings else None,
            "GoogleCloudSettings": self.google_cloud_settings.to_json() if self.google_cloud_settings else None,
        }


class RetentionPolicy:
    def __init__(self, disabled: bool = None, minimum_backup_age_to_keep: timedelta = None):
        self.disabled = disabled
        self.minimum_backup_age_to_keep = minimum_backup_age_to_keep

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> RetentionPolicy:
        return cls(json_dict["Disabled"], Utils.string_to_timedelta(json_dict["MinimumBackupAgeToKeep"]))

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "MinimumBackupAgeToKeep": Utils.timedelta_to_str(self.minimum_backup_age_to_keep),
        }


class PeriodicBackupConfiguration(BackupConfiguration):
    def __init__(
        self,
        backup_type: BackupType = None,
        snapshot_settings: SnapshotSettings = None,
        backup_encryption_settings: BackupEncryptionSettings = None,
        local_settings: LocalSettings = None,
        s3_settings: S3Settings = None,
        glacier_settings: GlacierSettings = None,
        azure_settings: AzureSettings = None,
        ftp_settings: FtpSettings = None,
        google_cloud_settings: GoogleCloudSettings = None,
        name: str = None,
        task_id: int = None,
        disabled: bool = None,
        mentor_node: str = None,
        pin_to_mentor_node: bool = None,
        retention_policy: RetentionPolicy = None,
        full_backup_frequency: str = None,
        incremental_backup_frequency: str = None,
    ):
        super().__init__(
            backup_type,
            snapshot_settings,
            backup_encryption_settings,
            local_settings,
            s3_settings,
            glacier_settings,
            azure_settings,
            ftp_settings,
            google_cloud_settings,
        )
        self.name = name
        self.task_id = task_id
        self.disabled = disabled
        self.mentor_node = mentor_node
        self.pin_to_mentor_node = pin_to_mentor_node
        self.retention_policy = retention_policy
        self.full_backup_frequency = full_backup_frequency
        self.incremental_backup_frequency = incremental_backup_frequency

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> PeriodicBackupConfiguration:
        backup_conf = BackupConfiguration.from_json(json_dict)
        return PeriodicBackupConfiguration(
            backup_conf.backup_type,
            backup_conf.snapshot_settings,
            backup_conf.backup_encryption_settings,
            backup_conf.local_settings,
            backup_conf.s3_settings,
            backup_conf.glacier_settings,
            backup_conf.azure_settings,
            backup_conf.ftp_settings,
            backup_conf.google_cloud_settings,
            json_dict["Name"],
            json_dict["TaskId"],
            json_dict["Disabled"],
            json_dict["MentorNode"],
            json_dict["PinToMentorNode"],
            RetentionPolicy.from_json(json_dict["RetentionPolicy"]) if json_dict["RetentionPolicy"] else None,
            json_dict["FullBackupFrequency"],
            json_dict["IncrementalBackupFrequency"],
        )

    def to_json(self) -> Dict[str, Any]:
        json_dict = super().to_json()
        json_dict.update(
            {
                "Name": self.name,
                "TaskId": self.task_id,
                "Disabled": self.disabled,
                "MentorNode": self.mentor_node,
                "PinToMentorNode": self.pin_to_mentor_node,
                "RetentionPolicy": self.retention_policy.to_json() if self.retention_policy else None,
                "FullBackupFrequency": self.full_backup_frequency,
                "IncrementalBackupFrequency": self.incremental_backup_frequency,
            }
        )

        return json_dict



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Callable, Optional

from ravendb.primitives import constants
from ravendb.documents.indexes.spatial.configuration import SpatialRelation, SpatialUnits
from ravendb.documents.session.tokens.misc import WhereOperator
from ravendb.documents.session.tokens.query_tokens.query_token import QueryToken
from ravendb.documents.session.tokens.query_tokens.definitions import ShapeToken, WhereToken


class DynamicSpatialField(ABC):
    def __init__(self, round_factor: float = 0):
        self.__round_factor = round_factor

    @property
    def round_factor(self) -> float:
        return self.__round_factor

    @abstractmethod
    def to_field(self, ensure_valid_field_name: Callable[[str, bool], str]) -> str:
        pass

    def round_to(self, factor: float) -> DynamicSpatialField:
        self.__round_factor = factor
        return self


class PointField(DynamicSpatialField):
    def __init__(self, latitude: str, longitude: str):
        super().__init__()
        self.latitude = latitude
        self.longitude = longitude

    def to_field(self, ensure_valid_field_name: Callable[[str, bool], str]) -> str:
        return (
            f"spatial.point({ensure_valid_field_name(self.latitude, False)}, "
            f"{ensure_valid_field_name(self.longitude, False)})"
        )


class WktField(DynamicSpatialField):
    def __init__(self, wkt: str):
        super().__init__()
        self.wkt = wkt

    def to_field(self, ensure_valid_field_name: Callable[[str, bool], str]) -> str:
        return f"spatial.wkt({ensure_valid_field_name(self.wkt, False)})"


class SpatialCriteria(ABC):
    def __init__(self, relation: SpatialRelation, distance_error_pct: float):
        self.__relation = relation
        self.__distance_error_pct = distance_error_pct

    @abstractmethod
    def _get_shape_token(self, add_query_parameter: Callable[[object], str]) -> ShapeToken:
        pass

    def to_query_token(self, field_name: str, add_query_parameter: Callable[[object], str]) -> QueryToken:
        shape_token = self._get_shape_token(add_query_parameter)

        if self.__relation == SpatialRelation.WITHIN:
            where_operator = WhereOperator.SPATIAL_WITHIN
        elif self.__relation == SpatialRelation.CONTAINS:
            where_operator = WhereOperator.SPATIAL_CONTAINS
        elif self.__relation == SpatialRelation.DISJOINT:
            where_operator = WhereOperator.SPATIAL_DISJOINT
        elif self.__relation == SpatialRelation.INTERSECTS:
            where_operator = WhereOperator.SPATIAL_INTERSECTS
        else:
            raise ValueError()

        return WhereToken.create(
            where_operator,
            field_name,
            None,
            WhereToken.WhereOptions(shape__distance=(shape_token, self.__distance_error_pct)),
        )


class WktCriteria(SpatialCriteria):
    def __init__(
        self, shape_wkt: str, relation: SpatialRelation, radius_units: SpatialUnits, distance_error_pct: float
    ):
        super().__init__(relation, distance_error_pct)
        self.__shape_wkt = shape_wkt
        self.__radius_units = radius_units

    def _get_shape_token(self, add_query_parameter: Callable[[object], str]) -> ShapeToken:
        return ShapeToken.wkt(add_query_parameter(self.__shape_wkt), self.__radius_units)


class CircleCriteria(SpatialCriteria):
    def __init__(
        self,
        radius: float,
        latitude: float,
        longitude: float,
        radius_units: SpatialUnits,
        relation: SpatialRelation,
        dist_error_percent: float,
    ):
        super().__init__(relation, dist_error_percent)

        self.__radius = radius
        self.__latitude = latitude
        self.__longitude = longitude
        self.__radius_units = radius_units

    def _get_shape_token(self, add_query_parameter: Callable[[object], str]) -> ShapeToken:
        return ShapeToken.circle(
            add_query_parameter(self.__radius),
            add_query_parameter(self.__latitude),
            add_query_parameter(self.__longitude),
            self.__radius_units,
        )


class SpatialCriteriaFactory:
    INSTANCE = None

    @classmethod
    def instance(cls) -> SpatialCriteriaFactory:
        if cls.INSTANCE is None:
            cls.INSTANCE = SpatialCriteriaFactory()

        return cls.INSTANCE

    def relates_to_shape(
        self,
        shape_wkt: str,
        relation: SpatialRelation,
        units: SpatialUnits = None,
        dist_error_percent: Optional[float] = constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT,
    ) -> SpatialCriteria:
        return WktCriteria(shape_wkt, relation, units, dist_error_percent)

    def intersects(
        self,
        shape_wkt: str,
        units: Optional[SpatialUnits] = None,
        dist_error_percent: Optional[float] = constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT,
    ) -> SpatialCriteria:
        return self.relates_to_shape(shape_wkt, SpatialRelation.INTERSECTS, units, dist_error_percent)

    def contains(
        self,
        shape_wkt: str,
        units: Optional[SpatialUnits] = None,
        dist_error_percent: Optional[float] = constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT,
    ) -> SpatialCriteria:
        return self.relates_to_shape(shape_wkt, SpatialRelation.CONTAINS, units, dist_error_percent)

    def disjoint(
        self,
        shape_wkt: str,
        units: Optional[SpatialUnits] = None,
        dist_error_percent: Optional[float] = constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT,
    ) -> SpatialCriteria:
        return self.relates_to_shape(shape_wkt, SpatialRelation.DISJOINT, units, dist_error_percent)

    def within(
        self,
        shape_wkt: str,
        units: Optional[SpatialUnits] = None,
        dist_error_percent: Optional[float] = constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT,
    ) -> SpatialCriteria:
        return self.relates_to_shape(shape_wkt, SpatialRelation.WITHIN, units, dist_error_percent)

    def within_radius(
        self,
        radius: float,
        latitude: float,
        longitude: float,
        radius_units: Optional[SpatialUnits] = None,
        dist_error_percent: Optional[float] = constants.Documents.Indexing.Spatial.DEFAULT_DISTANCE_ERROR_PCT,
    ) -> SpatialCriteria:
        return CircleCriteria(radius, latitude, longitude, radius_units, SpatialRelation.WITHIN, dist_error_percent)



def create_entity_with_mapper(dict_obj, mapper, object_type, convert_to_snake_case=None):
    """
    This method will create an entity from dict_obj and mapper
    In case convert_to_snake_case is empty will convert dict_obj keys to snake_case
    convert_to_snake_case can be dictionary with special words you can change ex. From -> from_date
    """
    from ravendb.tools.utils import Utils

    def parse_dict_rec(data):
        try:
            if not isinstance(data, dict):
                try:
                    for i in range(len(data)):
                        data[i] = Utils.initialize_object(
                            parse_dict_rec(data[i])[0],
                            object_type,
                            convert_to_snake_case,
                        )
                except TypeError:
                    return data, False

            for key in data:
                if isinstance(data[key], dict):
                    parse_dict_rec(data[key])
                    data[key] = mapper(key, data[key])
                    if key is None:
                        pass
                elif isinstance(data[key], (tuple, list)):
                    for item in data[key]:
                        parse_dict_rec(item)
                    data[key] = mapper(key, data[key])
                else:
                    mapper_result = mapper(key, data[key])
                    data[key] = mapper_result if mapper_result else data[key]
            return data, True
        except TypeError as e:
            raise TypeError("Can't parse to custom object", e)

    first_parsed, need_to_parse = parse_dict_rec(dict_obj)
    # After create a complete dict for our object we need to create the object with the object_type
    if first_parsed is not None and need_to_parse:
        return Utils.initialize_object(first_parsed, object_type, convert_to_snake_case)
    return first_parsed



from __future__ import annotations
from abc import abstractmethod
from typing import Generic, Optional, TypeVar, Union, Callable, Dict

from ravendb.documents.queries.facets.definitions import GenericRangeFacet, Facet, FacetBase, FacetAggregationField
from ravendb.documents.queries.facets.misc import FacetOptions, FacetAggregation
from ravendb.tools.utils import CaseInsensitiveSet

_T = TypeVar("_T")


class FacetOperations(Generic[_T]):
    @abstractmethod
    def with_display_name(self, display_name: str) -> FacetOperations[_T]:
        pass

    @abstractmethod
    def with_options(self, options: FacetOptions) -> FacetOperations[_T]:
        pass

    @abstractmethod
    def sum_on(self, path: str, display_name: Optional[str]) -> FacetOperations[_T]:
        pass

    @abstractmethod
    def min_on(self, path: str, display_name: Optional[str]) -> FacetOperations[_T]:
        pass

    @abstractmethod
    def max_on(self, path: str, display_name: Optional[str]) -> FacetOperations[_T]:
        pass

    @abstractmethod
    def average_on(self, path: str, display_name: Optional[str]) -> FacetOperations[_T]:
        pass


class FacetBuilder(Generic[_T], FacetOperations[_T]):
    _rql_keywords = CaseInsensitiveSet({"as", "select", "where", "load", "group", "order", "include", "update"})

    def __init__(self):
        self.__range: Union[None, GenericRangeFacet] = None
        self.__default: Union[None, Facet] = None

    def by_ranges(self, range_: RangeBuilder, *ranges: RangeBuilder) -> FacetOperations[_T]:
        if range_ is None:
            raise ValueError("range cannot be None")

        if self.__range is None:
            self.__range = GenericRangeFacet()

        self.__range.ranges.append(range_)

        if ranges:
            for p in ranges:
                self.__range.ranges.append(p)

        return self

    def by_field(self, field_name: str) -> FacetOperations[_T]:
        if self.__default is None:
            self.__default = Facet()

        if field_name in self._rql_keywords:
            field_name = f"'{field_name}'"

        self.__default.field_name = field_name
        return self

    def all_results(self) -> FacetOperations[_T]:
        if self.__default is None:
            self.__default = Facet()

        self.__default.field_name = None
        return self

    def with_options(self, options: FacetOptions) -> FacetOperations[_T]:
        self.facet.options = options
        return self

    def with_display_name(self, display_name: str) -> FacetOperations[_T]:
        self.facet.display_field_name = display_name
        return self

    def sum_on(self, path: str, display_name: Optional[str] = None) -> FacetOperations[_T]:
        aggregations_map = self.facet.aggregations
        aggregations = aggregations_map[FacetAggregation.SUM] = (
            aggregations_map[FacetAggregation.SUM] if FacetAggregation.SUM in aggregations_map else set()
        )

        aggregation_field = FacetAggregationField()
        aggregation_field.name = path
        aggregation_field.display_name = display_name

        aggregations.add(aggregation_field)

        return self

    def min_on(self, path: str, display_name: Optional[str] = None) -> FacetOperations[_T]:
        aggregations_map = self.facet.aggregations
        aggregations = aggregations_map[FacetAggregation.MIN] = (
            aggregations_map[FacetAggregation.MIN] if FacetAggregation.MIN in aggregations_map else set()
        )

        aggregation_field = FacetAggregationField()
        aggregation_field.name = path
        aggregation_field.display_name = display_name

        aggregations.add(aggregation_field)

        return self

    def max_on(self, path: str, display_name: Optional[str] = None) -> FacetOperations[_T]:
        aggregations_map = self.facet.aggregations
        aggregations = aggregations_map[FacetAggregation.MAX] = (
            aggregations_map[FacetAggregation.MAX] if FacetAggregation.MAX in aggregations_map else set()
        )

        aggregation_field = FacetAggregationField()
        aggregation_field.name = path
        aggregation_field.display_name = display_name

        aggregations.add(aggregation_field)

        return self

    def average_on(self, path: str, display_name: Optional[str] = None) -> FacetOperations[_T]:
        aggregations_map = self.facet.aggregations
        aggregations = aggregations_map[FacetAggregation.AVERAGE] = (
            aggregations_map[FacetAggregation.AVERAGE] if FacetAggregation.AVERAGE in aggregations_map else set()
        )

        aggregation_field = FacetAggregationField()
        aggregation_field.name = path
        aggregation_field.display_name = display_name

        aggregations.add(aggregation_field)

        return self

    @property
    def facet(self) -> FacetBase:
        return self.__default or self.__range


class RangeBuilder(Generic[_T]):
    def __init__(self, path: str):
        self.__less_bound: Union[None, _T] = None
        self.__greater_bound: Union[None, _T] = None
        self.__less_inclusive: Union[None, bool] = None
        self.__greater_inclusive: Union[None, bool] = None

        self.__less_set = False
        self.__greater_set = False

        self.__path = path

    @classmethod
    def from_json(cls, json_dict: Dict) -> RangeBuilder:
        builder = cls(json_dict["Path"])

        builder.__less_bound = json_dict["LessBound"]
        builder.__greater_bound = json_dict["GreaterBound"]
        builder.__less_inclusive = json_dict["LessInclusive"]
        builder.__greater_inclusive = json_dict["GreaterInclusive"]
        builder.__less_set = json_dict["LessSet"]
        builder.__greater_set = json_dict["GreaterSet"]

        return builder

    def to_json(self) -> Dict:
        return {
            "Path": self.__path,
            "LessBound": self.__less_bound,
            "GreaterBound": self.__greater_bound,
            "LessInclusive": self.__less_inclusive,
            "GreaterInclusive": self.__greater_inclusive,
            "LessSet": self.__less_set,
            "GreaterSet": self.__greater_set,
        }

    @classmethod
    def for_path(cls, path: str) -> RangeBuilder[_T]:
        return cls(path)

    def __create_clone(self) -> RangeBuilder[_T]:
        builder = RangeBuilder(self.__path)
        builder.__less_bound = self.__less_bound
        builder.__greater_bound = self.__greater_bound
        builder.__less_inclusive = self.__less_inclusive
        builder.__greater_inclusive = self.__greater_inclusive
        builder.__less_set = self.__less_set
        builder.__greater_set = self.__greater_set
        return builder

    def is_less_than(self, value: _T) -> RangeBuilder[_T]:
        if self.__less_set:
            raise RuntimeError("Less bound was already set")

        clone = self.__create_clone()
        clone.__less_bound = value
        clone.__less_inclusive = False
        clone.__less_set = True
        return clone

    def is_less_than_or_equal_to(self, value: _T) -> RangeBuilder[_T]:
        if self.__less_set:
            raise RuntimeError("Less bound was already set")

        clone = self.__create_clone()
        clone.__less_bound = value
        clone.__less_inclusive = True
        clone.__less_set = True
        return clone

    def is_greater_than(self, value: _T) -> RangeBuilder[_T]:
        if self.__greater_set:
            raise RuntimeError("Greater bound was already set")

        clone = self.__create_clone()
        clone.__greater_bound = value
        clone.__greater_inclusive = False
        clone.__greater_set = True
        return clone

    def is_greater_than_or_equal_to(self, value: _T) -> RangeBuilder[_T]:
        if self.__greater_set:
            raise RuntimeError("Greater bound was already set")

        clone = self.__create_clone()
        clone.__greater_bound = value
        clone.__greater_inclusive = True
        clone.__greater_set = True
        return clone

    def get_string_representation(self, add_query_parameter: Callable[[object], str]):
        less: str = None
        greater: str = None

        if not self.__less_set and not self.__greater_set:
            raise RuntimeError("Bounds were not set")

        if self.__less_set:
            less_param_name = add_query_parameter(self.__less_bound)
            less = self.__path + (" <= " if self.__less_inclusive else " < ") + "$" + less_param_name

        if self.__greater_set:
            greater_param_name = add_query_parameter(self.__greater_bound)
            greater = self.__path + (" >= " if self.__greater_inclusive else " > ") + "$" + greater_param_name

        if less is not None and greater is not None:
            return greater + " and " + less

        return less or greater



from __future__ import annotations
import datetime
import hashlib
import threading
from abc import ABC
from enum import Enum
from typing import Union, Optional, TYPE_CHECKING, List, Dict, Generic, TypeVar

from ravendb.http.misc import LoadBalanceBehavior, ReadBalanceBehavior

if TYPE_CHECKING:
    from ravendb.http.request_executor import RequestExecutor
    from ravendb.documents.queries.misc import Query
    from ravendb.documents.session.operations.query import QueryOperation
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )
    from ravendb.documents.store.definition import DocumentStore
    from ravendb.http.server_node import ServerNode

_T_Key = TypeVar("_T_Key")
_T_Value = TypeVar("_T_Value")


class TransactionMode(Enum):
    SINGLE_NODE = "single_node"
    CLUSTER_WIDE = "cluster_wide"

    def __str__(self):
        return self.value


class ForceRevisionStrategy(Enum):
    NONE = "None"
    BEFORE = "Before"

    def __str__(self):
        return self.value


class SessionInfo:
    _client_session_id_counter = threading.local()
    _client_session_id_counter.counter = 0

    def __init__(
        self, session: InMemoryDocumentSessionOperations, options: SessionOptions, document_store: DocumentStore
    ):
        if not document_store:
            raise ValueError("DocumentStore cannot be None")
        if not session:
            raise ValueError("Session cannot be None")
        self._session = session
        self._session_id: Union[None, int] = None
        self._session_id_used: Union[None, bool] = None
        self._load_balancer_context_seed = session.request_executor.conventions.load_balancer_context_seed
        self._can_use_load_balance_behavior = (
            session.conventions.load_balance_behavior == LoadBalanceBehavior.USE_SESSION_CONTEXT
            and session.conventions.load_balancer_per_session_context_selector is not None
        )
        self.document_store = document_store
        self.no_caching = options.no_caching

        self.last_cluster_transaction_index: Union[None, int] = None

    @property
    def can_use_load_balance_behavior(self) -> bool:
        return self._can_use_load_balance_behavior

    @property
    def session_id(self) -> int:
        if self._session_id is None:
            context = None
            selector = self._session.conventions.load_balancer_per_session_context_selector
            if selector is not None:
                context = selector(self._session.database_name)
            self._set_context_internal(context)
        self._session_id_used = True
        return self._session_id

    @property
    def context(self):
        # placeholder for convenient setter
        return None

    @context.setter
    def context(self, session_id: str):
        if not session_id or session_id.isspace():
            raise ValueError("Session key cannot be None or whitespace")

        self._set_context_internal(session_id)

        self._can_use_load_balance_behavior = (
            self._can_use_load_balance_behavior
            or self._session.conventions.load_balance_behavior == LoadBalanceBehavior.USE_SESSION_CONTEXT
        )

    def _set_context_internal(self, session_id: str) -> None:
        if self._session_id_used:
            raise RuntimeError(
                "Unable to set the session context after it has already been used. "
                "The session context can only be modified before it is utilized."
            )

        if session_id is None:
            v = self._client_session_id_counter.counter
            self._session_id = v
            v += 1
            self._client_session_id_counter = v
        else:
            self._session_id = int(hashlib.md5(session_id.encode("utf-8")).hexdigest(), 16)

    def increment_request_count(self) -> None:
        self._session.increment_requests_count()

    def get_current_session_node(self, request_executor: RequestExecutor) -> ServerNode:
        if request_executor.conventions.load_balance_behavior == LoadBalanceBehavior.USE_SESSION_CONTEXT:
            if self._can_use_load_balance_behavior:
                result = request_executor.get_node_by_session_id(self.session_id)
                return result.current_node

        read_balance_behavior = request_executor.conventions.read_balance_behavior

        if read_balance_behavior == ReadBalanceBehavior.NONE:
            result = request_executor.preferred_node
        elif read_balance_behavior == ReadBalanceBehavior.ROUND_ROBIN:
            result = request_executor.get_node_by_session_id(self.session_id)
        elif read_balance_behavior == ReadBalanceBehavior.FASTEST_NODE:
            result = request_executor.get_fastest_node()
        else:
            raise ValueError(f"Unsupported read balance behavior '{str(read_balance_behavior)}'")

        return result.current_node


class SessionOptions:
    def __init__(
        self,
        database: Optional[str] = None,
        no_tracking: Optional[bool] = None,
        no_caching: Optional[bool] = None,
        request_executor: Optional[RequestExecutor] = None,
        transaction_mode: Optional[TransactionMode] = None,
        disable_atomic_document_writes_in_cluster_wide_transaction: Optional[bool] = None,
    ):
        self.database = database
        self.no_tracking = no_tracking
        self.no_caching = no_caching
        self.request_executor = request_executor
        self.transaction_mode = transaction_mode
        self.disable_atomic_document_writes_in_cluster_wide_transaction = (
            disable_atomic_document_writes_in_cluster_wide_transaction
        )


class DocumentQueryCustomization:
    def __init__(self, query: Query):
        self.query = query
        self.query_operation: QueryOperation = None


class DocumentsChanges:
    class ChangeType(Enum):
        DOCUMENT_DELETED = "document_deleted"
        DOCUMENT_ADDED = "document_added"
        FIELD_CHANGED = "field_changed"
        NEW_FIELD = "new_field"
        REMOVED_FIELD = "removed_field"
        ARRAY_VALUE_CHANGED = "array_value_changed"
        ARRAY_VALUE_ADDED = "array_value_added"
        ARRAY_VALUE_REMOVED = "array_value_removed"

        def __str__(self):
            return self.value

    def __init__(
        self,
        field_old_value: object,
        field_new_value: object,
        change: ChangeType,
        field_name: str = None,
        field_path: str = None,
    ):
        self.field_old_value = field_old_value
        self.field_new_value = field_new_value
        self.change = change
        self.field_name = field_name
        self.field_path = field_path

    @property
    def field_full_name(self) -> str:
        return self.field_name if not self.field_path else f"{self.field_path}.{self.field_name}"


class ResponseTimeInformation:
    class ResponseTimeItem:
        def __init__(self, url: str = None, duration: datetime.timedelta = None):
            self.url = url
            self.duration = duration

    def __init__(
        self,
        total_server_duration: datetime.timedelta = datetime.timedelta.min,
        total_client_duration: datetime.timedelta = datetime.timedelta.min,
        duration_breakdown: List[ResponseTimeItem] = None,
    ):
        self.total_server_duration = total_server_duration
        self.total_client_duration = total_client_duration
        self.duration_breakdown = duration_breakdown if duration_breakdown is not None else []

    def compute_server_total(self):
        self.total_server_duration = sum(map(lambda x: x.duration, self.duration_breakdown))


class JavaScriptArray:
    def __init__(self, suffix: int, path_to_array: str):
        self.__suffix = suffix
        self.__path_to_array = path_to_array
        self.__arg_counter = 0
        self.__script_lines = []
        self.__parameters: Dict[str, object] = {}

    @property
    def script(self) -> str:
        return "\r".join(self.__script_lines)

    @property
    def parameters(self) -> Dict[str, object]:
        return self.__parameters

    def __get_next_argument_name(self) -> str:
        self.__arg_counter += 1
        return f"val_{self.__arg_counter - 1}_{self.__suffix}"

    def add(self, *u) -> JavaScriptArray:
        def __func(value) -> str:
            argument_name = self.__get_next_argument_name()
            self.__parameters[argument_name] = value
            return f"args.{argument_name}"

        args = ",".join(list(map(__func, u)))
        self.__script_lines.append(f"this.{self.__path_to_array}.push({args});")
        return self

    def remove_at(self, index: int) -> JavaScriptArray:
        argument_name = self.__get_next_argument_name()

        self.__script_lines.append(f"this.{self.__path_to_array}.splice(args.{argument_name}, 1);")
        self.__parameters[argument_name] = index

        return self


class JavaScriptMap(Generic[_T_Key, _T_Value]):
    def __init__(self, suffix: int, path_to_map: str):
        self._suffix = suffix
        self._path_to_map = path_to_map
        self._arg_counter = 0
        self._script_lines = []
        self._parameters: Dict[str, object] = {}

    @property
    def script(self) -> str:
        return "\r".join(self._script_lines)

    @property
    def parameters(self) -> Dict[str, object]:
        return self._parameters

    def _get_next_argument_name(self) -> str:
        self._arg_counter += 1
        return f"val_{self._arg_counter - 1}_{self._suffix}"

    def put(self, key: _T_Key, value: _T_Value) -> JavaScriptMap[_T_Key, _T_Value]:
        argument_name = self._get_next_argument_name()

        self._script_lines.append(f"this.{self._path_to_map}.{key} = args.{argument_name};")
        self.parameters[argument_name] = value
        return self

    def remove(self, key: _T_Key) -> JavaScriptMap[_T_Key, _T_Value]:
        self._script_lines.append(f"delete this.{self._path_to_map}.{key};")


class MethodCall(ABC):
    def __init__(self, args: List[object] = None, access_path: str = None):
        self.args = args
        self.access_path = access_path


class CmpXchg(MethodCall):
    @classmethod
    def value(cls, key: str) -> CmpXchg:
        cmp_xchg = cls()
        cmp_xchg.args = [key]

        return cmp_xchg


class OrderingType(Enum):
    STRING = 0
    LONG = " AS long"
    FLOAT = " AS double"
    ALPHA_NUMERIC = " AS alphaNumeric"

    def __str__(self):
        return self.value



from typing import Generic, TypeVar, Optional

_T = TypeVar("_T")


class Revision(Generic[_T]):
    def __init__(self, previous: Optional[_T] = None, current: Optional[_T] = None):
        self.previous = previous
        self.current = current



from __future__ import annotations
import json
from typing import Optional, Dict, Any

import requests

from ravendb import ServerNode
from ravendb.http.raven_command import RavenCommand
from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.operations.expiration.configuration import ExpirationConfiguration
from ravendb.documents.operations.definitions import MaintenanceOperation


class ConfigureExpirationOperationResult:
    def __init__(self, raft_command_index: int = None):
        self.raft_command_index = raft_command_index

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> ConfigureExpirationOperationResult:
        return cls(json_dict["RaftCommandIndex"])


class ConfigureExpirationOperation(MaintenanceOperation[ConfigureExpirationOperationResult]):
    def __init__(self, configuration: ExpirationConfiguration):
        self._configuration = configuration

    def get_command(self, conventions: DocumentConventions) -> RavenCommand:
        return self.ConfigureExpirationCommand(self._configuration)

    class ConfigureExpirationCommand(RavenCommand[ConfigureExpirationOperationResult]):
        def __init__(self, configuration: ExpirationConfiguration):
            super().__init__(ConfigureExpirationOperationResult)
            if configuration is None:
                raise ValueError("Configuration cannot be None")
            self._configuration = configuration

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/admin/expiration/config"

            request = requests.Request("POST", url)
            request.data = self._configuration.to_json()

            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = ConfigureExpirationOperationResult.from_json(json.loads(response))



import logging
from typing import Collection, TYPE_CHECKING, Optional, List, TypeVar

from ravendb.documents.indexes.definitions import IndexDefinition, IndexPriority, IndexState
from ravendb.documents.operations.indexes import PutIndexesOperation

if TYPE_CHECKING:
    from ravendb.documents.store.definition import DocumentStore
    from ravendb.documents.conventions import DocumentConventions
    from ravendb.documents.indexes.abstract_index_creation_tasks import AbstractIndexCreationTask


class IndexCreation:
    @staticmethod
    def create_indexes(
        indexes: Collection["AbstractIndexCreationTask"],
        store: "DocumentStore",
        conventions: Optional["DocumentConventions"] = None,
    ) -> None:
        if conventions is None:
            conventions = store.conventions

        try:
            indexes_to_add = IndexCreation.create_indexes_to_add(indexes, conventions)
            store.maintenance.send(PutIndexesOperation(*indexes_to_add))
        except Exception as e:
            logging.info("Could not create indexes in one shot (maybe using older version of RavenDB ?)", exc_info=e)
            for index in indexes:
                index.execute(store, conventions)

    @staticmethod
    def create_indexes_to_add(
        index_creation_tasks: "AbstractIndexCreationTask", conventions: "DocumentConventions"
    ) -> List[IndexDefinition]:
        def __map(x: "AbstractIndexCreationTask"):
            old_conventions = x.conventions
            try:
                x.conventions = conventions
                definition = x.create_index_definition()
                definition.name = x.index_name
                definition.priority = x.priority or IndexPriority.NORMAL
                definition.state = x.state or IndexState.NORMAL
                return definition
            finally:
                x.conventions = old_conventions

        return list(map(__map, index_creation_tasks))
